# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Post-drink Twitter annoy-o-matic.

To enable, you must:
- set `feature.twitter.enable` to TRUE in the Config table;
- set `contrib.twitter.login_name` and `contrib.twitter.login_pass` in Config to
  the user you want to use for tweets;
- add `pykeg.contrib.twitter` to the INSTALLED_APPS list (in the settings file);
- syncdb
"""

import random

from pykeg.core import backend
from pykeg.core import kb_app
from pykeg.core import models as core_models
from pykeg.core.net import kegnet
from pykeg.contrib.twitter import models

import twitter

TWITTER_NAME_UNKNOWN = 'Someone'
TWEET_MAX_LEN = 140

TWEET_TEMPLATE = (
    '%(tweet_to)s just poured %(ounces).1f oz'
    '%(keg_predicate)s.'
    '%(smug_remark)s %(url)s %(trailer)s'
)

class TwitterKegnetClient(kegnet.SimpleKegnetClient):
  def __init__(self, addr=None):
    kegnet.SimpleKegnetClient.__init__(self, addr)
    self._backend = backend.KegbotBackend()

  def onDrinkCreated(self, event):
    self._logger.info('Processing new drink %i for user %s' % (event.drink_id,
        event.username))

    # TODO(mikey): this should not be allowed to use the DB directly. Quick hack
    # for now..
    drink = core_models.Drink.objects.get(id=event.drink_id)

    tweet_to = TWITTER_NAME_UNKNOWN
    if drink.user:
      try:
        db_user = core_models.User.objects.get(username=drink.user.username)
        profile = db_user.get_profile()
      except (core_models.User.DoesNotExist, core_models.UserProfile.DoesNotExist):
        self._logger.warning('Profile for user %s does not exist' % (drink.user,))
        return
      try:
        twitter_link = models.UserTwitterLink.objects.get(user_profile=profile)
        tweet_to = '@' + twitter_link.twitter_name
        # TODO(mikey): determine if the user is following us; refuse to tweet if
        # not (so as to avoid tweeting @SomeRandomUser)
        # TODO(mikey): follow users who have established a link
      except models.UserTwitterLink.DoesNotExist:
        # Twitter name unknown: use kegbot username (with no '@')
        self._logger.warning('User %s has not enabled twitter link.' % (drink.user,))
        tweet_to = drink.user.username
    else:
      config = self._backend.GetConfig()
      if not config.getboolean('contrib.twitter.tweet_unknown'):
        self._logger.info('Tweeting for unknown users is disabled.')
        return

    tweet = self._GenerateTweet(drink, tweet_to)
    if not tweet:
      return

    try:
      tweet_log = self._DoTweet(tweet)
      models.DrinkTweetLog.objects.create(tweet_log=tweet_log, drink=drink)
    except Exception, e:
      self._logger.warning('Error posting tweet: %s' % (e,))
      return

    self._logger.info('Tweet posted: id=%s text="%s"' % (tweet_log.twitter_id,
        tweet_log.tweet))

  def _DoTweet(self, tweet):
    config = self._backend.GetConfig()
    twitter_name = config.get('contrib.twitter.login_name')
    twitter_pass = config.get('contrib.twitter.login_pass')

    # compress extra spaces
    tweet = ' '.join(tweet.split())

    if not twitter_name or not twitter_pass:
      self._logger.warning('No twitter login credentials')
      return

    api = twitter.Api(username=twitter_name, password=twitter_pass)
    api.SetXTwitterHeaders('Kegbot', 'http://kegbot.org/', '1.0')
    self._logger.info('Posting tweet [%i]: "%s"' % (len(tweet), tweet))
    status = api.PostUpdate(tweet)

    tweet_log = models.TweetLog.objects.create(twitter_id=str(status.id),
        tweet=tweet)
    return tweet_log

  def _GenerateTweet(self, drink, tweet_to):
    ounces = drink.Volume().ConvertTo.Ounce
    config = self._backend.GetConfig()

    try:
      min_ounces = float(config.get('contrib.twitter.min_oz'))
    except (TypeError, ValueError):
      min_ounces = 2.0

    if ounces < min_ounces:
      self._logger.info('Drink too small to care about; no tweet for you!')
      return

    smug_remark = models.DrinkRemark.GetRemarkForDrink(drink)
    if smug_remark:
      smug_remark = ' ' + smug_remark.remark
    else:
      smug_remark = ''
    trailer = ' #kegbot'

    keg_predicate = ''
    if drink.keg:
      keg_name = drink.keg.type.name
      if keg_name:
        keg_predicate = ' of %s' % drink.keg.type.name

    url = drink.ShortUrl()

    full_tweet = TWEET_TEMPLATE % vars()
    smug_remark = ''
    small_tweet = TWEET_TEMPLATE % vars()
    keg_predicate = ''
    tiny_tweet = TWEET_TEMPLATE % vars()

    self._logger.debug('Full tweet [%i]: %s' % (len(full_tweet), full_tweet))
    self._logger.debug('Small tweet [%i]: %s' % (len(small_tweet), small_tweet))
    self._logger.debug('Tiny tweet [%i]: %s' % (len(tiny_tweet), tiny_tweet))

    if len(full_tweet) < TWEET_MAX_LEN:
      return full_tweet
    elif len(small_tweet) < TWEET_MAX_LEN:
      return small_tweet
    elif len(tiny_weet) < TWEET_MAX_LEN:
      return tiny_tweet
    else:
      self._logger.error('All tweets were too large.')
      return ''


class TwitterApp(kb_app.App):
  def _MainLoop(self):
    self._client = TwitterKegnetClient()
    while not self._do_quit:
      self._client.serve_forever()

  def Quit(self):
    kb_app.App.Quit(self)
    self._client.stop()



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Wrapper module for database implementation."""

import datetime
import logging

from django.db.utils import DatabaseError

from pykeg.core import kb_common
from pykeg.core import config
from pykeg.core import models
from pykeg.core import protolib

from pykeg.web.api import common as krest_common
from pykeg.web.api import krest

class BackendError(Exception):
  """Base backend error exception."""

class Backend:
  """Abstract base Kegbot backend class.

  This class defines the interface that pykeg uses to talk to the kegbot
  backend.
  """

  def GetConfig(self):
    """Returns a KegbotConfig instance based on the current database values."""
    raise NotImplementedError

  def CreateNewUser(self, username, gender=kb_common.DEFAULT_NEW_USER_GENDER,
      weight=kb_common.DEFAULT_NEW_USER_WEIGHT):
    """Creates a new User instance.

    Args
      username: the unique username for the new User
      gender: the gender to assign the new user
      weight: the weight to assign the new user
    Returns
      the new User instance
    """
    raise NotImplementedError

  def GetAllTaps(self):
    """Returns all currently enabled taps."""
    raise NotImplementedError

  def RecordDrink(self, tap_name, ticks, volume_ml=None, username=None,
      pour_time=None, duration=None, auth_token=None):
    """Records a new drink with the given parameters."""
    raise NotImplementedError

  def LogSensorReading(self, sensor_name, temperature, when=None):
    """Records a new sensor reading."""
    raise NotImplementedError

  def GetAuthToken(self, auth_device, token_value):
    """Returns an AuthenticationToken instance."""
    raise NotImplementedError


class KegbotBackend(Backend):
  """Django models backed Backend."""

  def __init__(self, sitename='default', site=None):
    self._logger = logging.getLogger('backend')
    self._config = config.KegbotConfig(self._GetConfigDict())
    if site:
      self._site = site
    else:
      self._site = models.KegbotSite.objects.get(name=sitename)

  def _GetConfigDict(self):
    try:
      ret = {}
      for row in models.Config.objects.all():
        ret[row.key] = row.value
      return ret
    except DatabaseError, e:
      raise BackendError, e

  def _GetTapFromName(self, tap_name):
    try:
      return models.KegTap.objects.get(site=self._site, meter_name=tap_name)
    except models.KegTap.DoesNotExist:
      return None

  def _GetKegForTapName(self, tap_name):
    tap = self._GetTapFromName(tap_name)
    if tap and tap.current_keg and tap.current_keg.status == 'online':
      return tap.current_keg
    return None

  def _GetSensorFromName(self, name, autocreate=True):
    try:
      return models.ThermoSensor.objects.get(site=self._site, raw_name=name)
    except models.ThermoSensor.DoesNotExist:
      if autocreate:
        sensor = models.ThermoSensor(site=self._site, raw_name=name, nice_name=name)
        sensor.save()
        return sensor
      else:
        return None

  def _GetUserObjFromUsername(self, username):
    try:
      return models.User.objects.get(username=username)
    except models.User.DoesNotExist:
      return None

  def GetConfig(self):
    return self._config

  def CreateNewUser(self, username, gender=kb_common.DEFAULT_NEW_USER_GENDER,
      weight=kb_common.DEFAULT_NEW_USER_WEIGHT):
    u = models.User(username=username)
    u.save()
    p = u.get_profile()
    p.gender = gender
    p.weight = weight
    p.save()
    return protolib.ToProto(u)

  def CreateAuthToken(self, auth_device, token_value, username=None):
    token = models.AuthenticationToken.objects.create(
        site=self._site, auth_device=auth_device, token_value=token_value)
    if username:
      user = self._GetUserObjFromUsername(username)
      token.user = user
    token.save()
    return protolib.ToProto(token)

  def GetAllTaps(self):
    return protolib.ToProto(list(models.KegTap.objects.all()))

  def RecordDrink(self, tap_name, ticks, volume_ml=None, username=None,
      pour_time=None, duration=None, auth_token=None):

    tap = self._GetTapFromName(tap_name)
    if not tap:
      self._logger.warning('No tap found on site %s with name %s' % (self._site,
          tap_name))
      return None

    d = models.Drink(ticks=ticks, site=self._site)

    if volume_ml is not None:
      d.volume_ml = volume_ml
    else:
      d.volume_ml = float(ticks) * tap.ml_per_tick

    if username:
      d.user = self._GetUserObjFromUsername(username)

    if not pour_time:
      pour_time = datetime.datetime.now()
    d.endtime = pour_time

    if duration:
      d.starttime = pour_time - datetime.timedelta(seconds=duration)
      d.duration = duration
    else:
      d.starttime = pour_time
      d.duration = 0

    # Look up the current keg.
    if tap.current_keg and tap.current_keg.status == 'online':
      d.keg = tap.current_keg

    d.auth_token = auth_token
    models.DrinkingSession.AssignSessionForDrink(d)
    d.save()

    d.PostProcess()

    return protolib.ToProto(d, full=False)

  def LogSensorReading(self, sensor_name, temperature, when=None):
    if not when:
      when = datetime.datetime.now()

    # The maximum resolution of ThermoSensor records is 1 minute.  Round the
    # time down to the nearest minute; if a record already exists for this time,
    # replace it.
    when = when.replace(second=0, microsecond=0)

    # If the temperature is out of bounds, reject it.
    min_val = kb_common.THERMO_SENSOR_RANGE[0]
    max_val = kb_common.THERMO_SENSOR_RANGE[1]
    if temperature < min_val or temperature > max_val:
      raise ValueError, 'Temperature out of bounds'

    sensor = self._GetSensorFromName(sensor_name)
    defaults = {
        'temp': temperature,
    }
    record, created = models.Thermolog.objects.get_or_create(site=self._site,
        sensor=sensor, time=when, defaults=defaults)
    record.temp = temperature
    record.save()
    return protolib.ToProto(record)

  def GetAuthToken(self, auth_device, token_value):
    try:
      return protolib.ToProto(models.AuthenticationToken.objects.get(
          site=self._site, auth_device=auth_device, token_value=token_value))
    except models.AuthenticationToken.DoesNotExist:
      return None


class WebBackend(Backend):
  def __init__(self, base_url, api_auth_token):
    self._logger = logging.getLogger('api-backend')
    self._base_url = base_url
    self._client = krest.KrestClient(base_url=base_url,
        api_auth_token=api_auth_token)

  def GetConfig(self):
    raise NotImplementedError

  def CreateNewUser(self, username, gender=kb_common.DEFAULT_NEW_USER_GENDER,
      weight=kb_common.DEFAULT_NEW_USER_WEIGHT):
    raise NotImplementedError

  def CreateAuthToken(self, auth_device, token_value, username=None):
    raise NotImplementedError

  def GetAllTaps(self):
    return self._client.TapStatus()

  def RecordDrink(self, tap_name, ticks, volume_ml=None, username=None,
      pour_time=None, duration=None, auth_token=None):
    return self._client.RecordDrink(tap_name, ticks, volume_ml, username,
        pour_time, duration, auth_token)

  def LogSensorReading(self, sensor_name, temperature, when=None):
    # If the temperature is out of bounds, reject it.
    min_val = kb_common.THERMO_SENSOR_RANGE[0]
    max_val = kb_common.THERMO_SENSOR_RANGE[1]
    if temperature < min_val or temperature > max_val:
      raise ValueError, 'Temperature out of bounds'

    try:
      return self._client.LogSensorReading(sensor_name, temperature, when)
    except krest_common.NotFoundError:
      self._logger.warning('No sensor on backend named "%s"' % (sensor_name,))
      return None

  def GetAuthToken(self, auth_device, token_value):
    return self._client.GetToken(auth_device, token_value)



from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

import datetime

from pykeg.core import backend
from pykeg.core import defaults
from pykeg.core import models
from pykeg.core import units
from django.contrib.auth.models import User


class Command(BaseCommand):
  help = u'Generate test data in a new kegbot database.'
  args = '<none>'

  def handle(self, *args, **options):
    if len(args) != 0:
      raise CommandError('No arguments required')

    defaults.gentestdata()



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to generate cached statistics from drinks."""

import inspect
import itertools

def stat(name, default=None):
  def decorate(f):
    setattr(f, 'statname', name)
    setattr(f, 'statdefault', default)
    return f
  return decorate

class StatsBuilder(object):

  def __init__(self, prev_stats=None):
    if prev_stats is None:
      prev_stats = {}
    self._prev_stats = prev_stats

  def AllStats(self):
    for name, method in inspect.getmembers(self, inspect.ismethod):
      if hasattr(method, 'statname'):
        yield (method.statname, method.statdefault, method)

  def Build(self, obj, prev_stats=None, prev_objs=None):
    if prev_objs is None:
      prev_objs = self._PrevObjs(obj).order_by('endtime')

    stats = self._prev_stats
    for drink in prev_objs:
      for statname, statdefault, fn in self.AllStats():
        previous_value = stats.get(statname, statdefault)
        stats[statname] = fn(drink, previous_value)
    return stats


class BaseStatsBuilder(StatsBuilder):
  """Builder which generates a variety of stats from object information."""

  def _PrevObjs(self, obj):
    raise NotImplementedError

  @stat('total_volume', default=0)
  def TotalVolume(self, drink, prev):
    """Updates the total volume."""
    return float(drink.volume_ml) + prev

  @stat('total_count', default=0)
  def TotalPours(self, drink, prev):
    return prev + 1

  @stat('volume_avg')
  def AverageVolume(self, drink, prev):
    if prev is None:
      prev = {'count': 0, 'average': 0}
    res = prev.copy()
    vol = float(drink.Volume())
    res['average'] = (res['average']*res['count'] + vol) / (res['count'] + 1)
    res['count'] = res['count'] + 1
    return res

  @stat('volume_max')
  def LargestVolume(self, drink, prev):
    prev = {}
    if not prev:
      prev = {'volume': 0, 'id': 0}
    vol = float(drink.Volume())
    if prev.get('volume', 0) >= vol:
      return prev
    return {'volume': vol, 'id': drink.id}

  @stat('volume_min')
  def SmallestPour(self, drink, prev):
    prev = {}
    if not prev:
      prev = {'volume': 0, 'id': 0}
    vol = float(drink.Volume())
    if prev.get('volume', 0) < vol:
      return prev
    return {'volume': vol, 'id': drink.id}

  @stat('volume_by_day_of_week')
  def VolumeByDayOfweek(self, drink, prev):
    volmap = prev
    if not volmap:
      volmap = dict((i, 0) for i in xrange(7))
    # Note: uses the session's starttime, rather than the drink's. This causes
    # late-night sessions to be reported for the day on which they were started.
    weekday = drink.session.starttime.weekday()
    volmap[weekday] += float(drink.Volume())
    return volmap

  @stat('volume_by_drinker')
  def VolumeByDrinker(self, drink, prev):
    volmap = prev
    if not volmap:
      volmap = {}
    if drink.user:
      username = drink.user.username
    else:
      username = None
    volmap[username] = volmap.get(username, 0) + float(drink.Volume())
    return volmap

  @stat('ids')
  def Ids(self, drink, prev):
    if prev is None:
      prev = []
    drink_id = drink.id
    if drink_id not in prev:
      prev.append(drink_id)
    return prev


class DrinkerStatsBuilder(BaseStatsBuilder):
  """Builder of user-specific stats by drink."""
  REVISION = 3

  def _PrevObjs(self, drink):
    qs = drink.user.drinks.valid().filter(endtime__lt=drink.endtime)
    qs = qs.order_by('endtime')
    return qs

  #@stat('volume-per-user-session')
  def VolumePerSession(self, drink, prev):
    if not prev:
      prev = []
    sess = drink.GetSession()
    user_sess = drink.user.session_parts.filter(session=sess)[0]
    res = [itertools.ifilter(lambda x: x[0] != user_sess.id, prev)]
    res.append((user_sess.id, float(user_sess.Volume())))


class KegStatsBuilder(BaseStatsBuilder):
  """Builder of keg-specific stats."""
  REVISION = 3

  def _PrevObjs(self, drink):
    qs = drink.keg.drinks.valid().filter(endtime__lt=drink.endtime)
    qs = qs.order_by('endtime')
    return qs

  @stat('drinkers')
  def Drinkers(self, drink, prev):
    if not prev:
      prev = []
    if drink.user and drink.user.id not in prev:
      prev.append(drink.user.id)
    return prev


def main():
  from pykeg.core import models

  for user in models.User.objects.all():
    last_drink = user.drinks.all().order_by('-endtime')
    if not last_drink:
      continue
    last_drink = last_drink[0]
    builder = DrinkerStatsBuilder()
    stats = builder.Build(last_drink)
    print '-'*72
    print 'stats for %s' % user
    for k, v in stats.iteritems():
      print '   %s: %s' % (k, v)
    print ''

if __name__ == '__main__':
  import cProfile
  command = """main()"""
  cProfile.runctx( command, globals(), locals(), filename="stats.profile" )




from django.conf import settings

import facebook

def profile_for_user(user):
  profile = user.facebookprofile_set.all()
  if not profile:
    return None
  return profile[0]

def session_for_user(user):
  profile = profile_for_user(user)
  if not profile:
    return None
  session = profile.session.all()
  if not session:
    return None
  return session[0]

def settings_for_user(user):
  profile = profile_for_user(user)
  if not profile:
    return None
  settings = profile.settings.all()
  if not settings:
    return None # should never happen, due to post-save signal
  return settings[0]

def stream_publish(user, **kwargs):
  session = session_for_user(user)
  if not session:
    raise ValueError, "No session."
  fb = facebook.Facebook(settings.FACEBOOK_API_KEY,
      settings.FACEBOOK_SECRET_KEY)
  fb.session_key = session.session_id
  fb.session_key_expires = 0
  return fb.stream.publish(**kwargs)




# Copyright 2008 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Python interfaces to a Kegboard device."""

import cStringIO
import logging
import struct
import string

import gflags

from pykeg.core import util
from pykeg.hw.kegboard import crc16

FLAGS = gflags.FLAGS

gflags.DEFINE_string('kegboard_device', '/dev/ttyUSB0',
    'An explicit device file (eg /dev/ttyUSB0) on which to listen for kegboard '
    'packets.')

gflags.DEFINE_integer('kegboard_speed', 115200,
    'Baud rate of device at --kegboard_device')

KBSP_PREFIX = "KBSP v1:"
KBSP_PAYLOAD_MAXLEN = 112
KBSP_TRAILER = "\r\n"

class KegboardError(Exception):
  """Generic error with Kegboard"""

class MessageError(KegboardError):
  """Generic error in Message class"""

class UnknownMessageError(MessageError):
  """Message id is not known"""

class FramingError(KegboardError):
  """Problem synchronizing with serial stream"""


### Fields

class Field(util.BaseField):
  def __init__(self, tagnum):
    self.tagnum = tagnum

  def ToBytes(self, value):
    raise NotImplementedError

  def ToString(self, value):
    return str(value)


class StructField(Field):
  _STRUCT_FORMAT = '<'
  def __init__(self, tagnum):
    Field.__init__(self, tagnum)
    self._packed_size = struct.calcsize(self._STRUCT_FORMAT)

  def ParseValue(self, value):
    if len(value) != self._packed_size:
      raise ValueError, "Bad length, must be exactly %i bytes" % (self._packed_size,)
    return struct.unpack(self._STRUCT_FORMAT, value)[0]

  def ToBytes(self, value):
    return struct.pack(self._STRUCT_FORMAT, value)


class Uint8Field(StructField):
  _STRUCT_FORMAT = 'B'


class Uint16Field(StructField):
  _STRUCT_FORMAT = '<H'


class Uint32Field(StructField):
  _STRUCT_FORMAT = '<I'


class Uint64Field(StructField):
  _STRUCT_FORMAT = '<Q'


class OnewireIdField(Uint64Field):
  def ToString(self, value):
    return "0x%x" % (value,)


class StringField(Field):
  def ParseValue(self, bytes):
    return bytes.strip('\x00')

  def ToBytes(self, value):
    return str(value) + '\x00'


class OutputField(Uint16Field):
  def ParseValue(self, bytes):
    if bytes.strip('\x00'):
      return 1
    else:
      return 0

  def ToString(self, value):
    if value:
      return 'on'
    else:
      return 'off'


class TempField(Uint32Field):
  def ParseValue(self, value):
    value = Uint32Field.ParseValue(self, value)
    value /= 1000000.0
    return value


### Message types

class Message(util.BaseMessage):
  def __init__(self, initial=None, bytes=None, payload_bytes=None, **kwargs):
    util.BaseMessage.__init__(self, initial, **kwargs)
    self._tag_to_field = {}
    for field in self._fields.itervalues():
      self._tag_to_field[field.tagnum] = field
    if bytes is not None:
      self.UnpackFromBytes(bytes)
    if payload_bytes is not None:
      self.UnpackFromPayload(payload_bytes)

  def UnpackFromBytes(self, bytes):
    if len(bytes) < 16:
      raise ValueError, "Not enough bytes"

    header = bytes[:12]
    payload = bytes[12:-4]
    trailer = bytes[-4:]
    crcd_bytes = bytes[:-2]

    prefix, message_id, message_len = struct.unpack('<8sHH', header)

    if len(payload) != message_len:
      raise ValueError, "Payload size does not match tag"

    # TODO(mikey): assert checked_crc == 0
    checked_crc = crc16.crc16_ccitt(crcd_bytes)
    self.UnpackFromPayload(payload)

  def UnpackFromPayload(self, payload):
    pos = 0
    payload_len = len(payload)
    while (pos + 2) <= payload_len:
      fieldlen = self._ParseField(payload[pos:])
      pos += 2 + fieldlen

  def _ParseField(self, field_bytes):
    tag, length = struct.unpack('<BB', field_bytes[:2])
    data = field_bytes[2:2+length]
    # If field number is known, set its value. (Ignore it otherwise.)
    field = self._tag_to_field.get(tag)
    if field:
      setattr(self, field.name, data)
    return length

  def ToBytes(self):
    payload = cStringIO.StringIO()
    for field_name, field in self._fields.iteritems():
      field_bytes = field.ToBytes(self._values[field_name])
      payload.write(struct.pack('<BB', field.tagnum, len(field_bytes)))
      payload.write(field_bytes)

    payload_str = payload.getvalue()
    payload.close()

    out = cStringIO.StringIO()
    out.write(KBSP_PREFIX)
    out.write(struct.pack('<HH', self.MESSAGE_ID, len(payload_str)))
    out.write(payload_str)
    crc = crc16.crc16_ccitt(out.getvalue())
    out.write(struct.pack('<H', crc))
    out.write('\r\n')
    return out.getvalue()

class HelloMessage(Message):
  MESSAGE_ID = 0x01
  firmware_version = Uint16Field(0x01)


class ConfigurationMessage(Message):
  MESSAGE_ID = 0x02
  board_name = StringField(0x01)
  baud_rate = Uint16Field(0x02)
  update_interval = Uint16Field(0x03)


class MeterStatusMessage(Message):
  MESSAGE_ID = 0x10
  meter_name = StringField(0x01)
  meter_reading = Uint32Field(0x02)


class TemperatureReadingMessage(Message):
  MESSAGE_ID = 0x11
  sensor_name = StringField(0x01)
  sensor_reading = TempField(0x02)


class OutputStatusMessage(Message):
  MESSAGE_ID = 0x12
  output_name = StringField(0x01)
  output_reading = OutputField(0x02)


class OnewirePresenceMessage(Message):
  MESSAGE_ID = 0x13
  device_id = OnewireIdField(0x01)
  status = Uint8Field(0x02)


class PingCommand(Message):
  MESSAGE_ID = 0x81


MESSAGE_ID_TO_CLASS = {}
for cls in Message.__subclasses__():
  idnum = cls.MESSAGE_ID
  if idnum in MESSAGE_ID_TO_CLASS:
    raise RuntimeError, "More than one message for id: %i" % (idnum,)
  MESSAGE_ID_TO_CLASS[idnum] = cls


def GetMessageForBytes(bytes):
  if len(bytes) < 6:
    raise ValueError, "Not enough bytes"
  prefix, message_id = struct.unpack('<8sH', bytes[:10])
  cls = MESSAGE_ID_TO_CLASS.get(message_id)
  if not cls:
    raise UnknownMessageError
  return cls(bytes=bytes)


def GetMessageById(message_id, payload_bytes=None):
  cls = MESSAGE_ID_TO_CLASS.get(message_id)
  if not cls:
    raise UnknownMessageError
  return cls(payload_bytes=payload_bytes)


class KegboardReader(object):
  def __init__(self, fd):
    self._logger = logging.getLogger('kegboard-reader')
    self._fd = fd

  def WriteMessage(self, message):
    if not isinstance(message, Message):
      raise ValueError, "WriteMessage must be called with a Message instance"
    self._fd.write(message.ToBytes())

  def GetNextMessage(self):
    header_pos = 0
    header_len = len(KBSP_PREFIX)
    while True:
      # Find the prefix, re-aligning the messages as needed.
      logged_frame_error = False
      while header_pos < header_len:
        byte = self._fd.read(1)
        if byte == KBSP_PREFIX[header_pos]:
          header_pos += 1
          if logged_frame_error:
            self._logger.info('Packet framing fixed.')
            logged_frame_error = False
        else:
          if not logged_frame_error:
            self._logger.info('Packet framing broken (found "%s", expected "%s"'
                              '); reframing.' % (byte, KBSP_PREFIX[header_pos]))
            logged_frame_error = True
          if byte == KBSP_PREFIX[0]:
            header_pos = 0
          else:
            header_pos = 0

      # Read message type and message length.
      header_bytes = self._fd.read(4)
      message_id, message_len = struct.unpack('<HH', header_bytes)
      if message_len > KBSP_PAYLOAD_MAXLEN:
        self._logger.warning('Bogus message length (%i), skipping message' %
                             message_len)
        continue

      # Read payload and trailer.
      payload = self._fd.read(message_len)
      crc = self._fd.read(2)
      trailer = self._fd.read(2)

      if trailer != KBSP_TRAILER:
        self._logger.warning('Bad trailer (%s), skipping message' %
                             repr(trailer))
        # Bogus trailer; start over.
        continue

      return GetMessageById(message_id, payload)



import asyncore
import datetime
import Queue
import time

import gflags

from pykeg.core import kb_common
from pykeg.core import util
from pykeg.core.net import kegnet
from pykeg.core.net import kegnet_pb2

FLAGS = gflags.FLAGS

### Base kegbot thread class

class CoreThread(util.KegbotThread):
  """ Convenience wrapper around a threading.Thread """
  def __init__(self, kb_env, name):
    util.KegbotThread.__init__(self, name)
    self._kb_env = kb_env

  ### Event listener methods
  def PostEvent(self, event):
    if isinstance(event, kegnet_pb2.QuitEvent):
      self._logger.info('got quit event, quitting')
      self.Quit()


class WatchdogThread(CoreThread):
  """Monitors all threads in _kb_env for crashes."""

  def ThreadMain(self):
    fault_detected = False
    while not self._quit:
      if not fault_detected:
        for thr in self._kb_env.GetThreads():
          if not thr.hasStarted():
            continue
          if not self._quit and not thr.isAlive():
            self._logger.error('Thread %s died unexpectedly' % thr.getName())
            event = kegnet_pb2.QuitEvent()
            self._kb_env.GetEventHub().PublishEvent(event)
            fault_detected = True
            break
      time.sleep(1.0)


class EventHubServiceThread(CoreThread):
  """Handles all event dispatches for the event hub."""

  def ThreadMain(self):
    hub = self._kb_env.GetEventHub()
    while not self._quit:
      hub.DispatchNextEvent(timeout=0.5)


class HeartbeatThread(CoreThread):
  """Generates periodic events."""

  def ThreadMain(self):
    hub = self._kb_env.GetEventHub()
    seconds = 0
    while not self._quit:
      time.sleep(1.0)
      seconds += 1
      event = kegnet_pb2.HeartbeatSecondEvent()
      hub.PublishEvent(event)
      if (seconds % 60) == 0:
        event = kegnet_pb2.HeartbeatMinuteEvent()
        hub.PublishEvent(event)
      if (seconds % 3600) == 0:
        event = kegnet_pb2.HeartbeatHourEvent()
        hub.PublishEvent(event)
        seconds = 0


class AlarmManagerThread(CoreThread):

  def ThreadMain(self):
    am = self._kb_env.GetAlarmManager()
    while not self._quit:
      alarm = am.WaitForNextAlarm(1.0)
      if alarm is not None:
        self._logger.info('firing alarm: %s' % alarm)
        event = alarm.event()
        self._kb_env.GetEventHub().PublishEvent(event)


class EventHandlerThread(CoreThread):
  """ Basic event handling thread. """
  def __init__(self, kb_env, name):
    CoreThread.__init__(self, kb_env, name)
    self._event_queue = Queue.Queue()
    self._event_handlers = set()
    self._all_event_map = {}

  def AddEventHandler(self, event_handler):
    self._event_handlers.add(event_handler)
    self._RefreshEventMap()

  def _RefreshEventMap(self):
    for svc in self._event_handlers:
      for event_type, callback_list in svc.GetEventHandlers().iteritems():
        if self._all_event_map.get(event_type) is None:
          self._all_event_map[event_type] = set()
        for cb in callback_list:
          self._all_event_map[event_type].add(cb)

  def GetStatus(self):
    lines = []
    for handler in self._event_handlers:
      handler_lines = handler.GetStatus()
      if handler_lines:
        name = handler.__class__.__name__
        lines.append('>> %s' % name)
        for h_line in handler_lines:
          lines.append('  %s' % h_line)
        lines.append('')
    return lines

  def ThreadMain(self):
    while not self._quit:
      self._Step(timeout=0.5)

  def _Step(self, timeout=0.5):
    event = self._WaitForEvent(timeout)
    if event is not None:
      self._ProcessEvent(event)
    return event

  def PostEvent(self, event):
    self._event_queue.put(event)

  def _GetCallbacksForEvent(self, event):
    return self._all_event_map.get(event.__class__, tuple())

  def _WaitForEvent(self, timeout=0.5):
    """ Block until an event is posted, then process it """
    try:
      ev = self._event_queue.get(timeout=timeout)
      return ev
    except Queue.Empty:
      return None

  def _ProcessEvent(self, event):
    """ Execute the event callback associated with the event, if present. """
    self._logger.debug('Processing event: %s' % event.DESCRIPTOR.name)
    if isinstance(event, kegnet_pb2.QuitEvent):
      self._logger.info('got quit event, quitting')
      self.Quit()
      return
    callbacks = self._GetCallbacksForEvent(event)
    for cb in callbacks:
      cb(event)

  def _FlushEvents(self):
    """ Process all events in the Queue immediately """
    while True:
      ev = self._Step(timeout=0.5)
      if ev is None:
        break


### Service threads
class NetProtocolThread(CoreThread):
  def ThreadMain(self):
    self._logger.info("network thread started")
    server = self._kb_env.GetKegnetServer()
    server.StartServer()
    while not self._quit:
      asyncore.loop(timeout=0.5, count=1)
    server.StopServer()



#!/usr/bin/env python
#
# Copyright 2008 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Chart functions for use by the views module."""

# This module requires pygooglechart. If pygooglechart is unavailable, all
# functions that can return a chart will instead return None.
try:
  import pygooglechart
  USE_GOOGLE_CHARTS = True
except ImportError:
  USE_GOOGLE_CHARTS = False

def volume_chart(vol_by_user):
  """Returns a PieChart2D representing per-user shares.

  Args
    vol_by_user - tuple of (volume in ounces, User)
  Returns
    PieChart2D, or None
  """
  if not USE_GOOGLE_CHARTS:
    return None
  chart = pygooglechart.PieChart2D(300, 125)
  data = []
  labels = []

  for amount, user in vol_by_user[:10]:
    ounces = amount.ConvertTo.Ounce
    data.append(ounces)
    labels.append("%s (%i oz)" % (user.username, ounces))

  other_vol = 0
  for amount, user in vol_by_user[10:]:
    ounces = amount.ConvertTo.Ounce
    other_vol += ounces
  if other_vol:
    data.append(other_vol)
    labels.append("all others (%i oz)" % other_vol)

  chart.add_data(data)
  chart.set_pie_labels(labels)
  return chart




# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellanous configuration from the database."""

class KegbotConfig:
  def __init__(self, configdict):
    self._configdict = configdict

  def get(self, option, default=None):
    return self._configdict.get(option, default)

  def getboolean(self, option):
    val = self.get(option, 'false')
    if val.lower() == 'true' or val == '1':
      return True
    return False

  def IsFeatureEnabled(self, feature_name):
    key = '.'.join(('feature', feature_name, 'enable'))
    return self.getboolean(key)




import datetime
from django.db import models
from django.db.models.signals import post_save
from socialregistration import models as sr_models

PRIVACY_CHOICES = (
  ('EVERYONE', 'Everyone'),
  ('ALL_FRIENDS', 'Friends'),
  ('FRIENDS_OF_FRIENDS', 'Friends of Friends'),
  ('NETWORK_FRIENDS', 'Networks and Friends'),
  #('CUSTOM', 'Custom permissions'),
)

class FacebookSession(models.Model):
  """Stores the session id for a user."""
  profile = models.ForeignKey(sr_models.FacebookProfile, unique=True,
      related_name='session')
  session_id = models.CharField(max_length=255, blank=False, null=False)
  updated = models.DateTimeField(default=datetime.datetime.now)

  @classmethod
  def get_session(cls, request):
    if not hasattr(request, 'facebook'):
      raise ValueError, "no facebook"
      return None
    fb = request.facebook
    if not fb.uid:
      raise ValueError, "no uid"
      return None

    profile = sr_models.FacebookProfile.objects.get(uid=fb.uid)
    if not profile:
      raise ValueError, "no profile"
      return None

    session, new = FacebookSession.objects.get_or_create(profile=profile)
    if new or session.session_id != fb.session_key:
      session.session_id = fb.session_key
      session.save()

  def add_permission(self, perm):
    qs = self.profile.permission_set.filter(permission=perm)
    if not qs.count():
      perm = FacebookPermission(profile=self.profile, permission=perm)
      perm.save()

  def rm_permission(self, perm):
    qs = self.profile.permission_set.filter(permission=perm)
    if qs.count():
      qs.delete()

def profile_post_save(sender, instance, **kwargs):
  """Create default settings on new profile."""
  settings, new = FacebookSettings.objects.get_or_create(
      profile=instance)
post_save.connect(profile_post_save, sender=sr_models.FacebookProfile)

class FacebookPermission(models.Model):
  """Records a granted permission."""
  profile = models.ForeignKey(sr_models.FacebookProfile, unique=True,
      related_name='permission_set')
  permission = models.CharField(max_length=255, blank=False, null=False,
      unique=True)

class FacebookSettings(models.Model):
  profile = models.ForeignKey(sr_models.FacebookProfile, unique=True,
      related_name='settings')

  # stream.publish stuff
  # http://wiki.developers.facebook.com/index.php/Stream.publish
  publish_events = models.BooleanField(default=False,
      help_text='Post each drink to your wall.')
  include_link = models.BooleanField(default=False,
      help_text='Add a link to this kegbot when publishing to wall.')
  publish_status = models.BooleanField(default=False,
      help_text='Update status on start of a new drinking session.')

  privacy = models.CharField(max_length=64, choices=PRIVACY_CHOICES,
      default='ALL_FRIENDS',
      help_text='Privacy setting for drink posts.')





from pykeg.core import util

### Drink-related constants

# Don't record teeny drinks
MIN_VOLUME_TO_RECORD = 10

# Don't show small drinks
MIN_VOLUME_TO_DISPLAY = 10

# Minimum session size, in milliliters, to display
MIN_SESSION_VOLUME_DISPLAY_ML = 177  # 6 ounces

# Idle seconds allowed before a user gets booted for inactivity
FLOW_IDLE_TIMEOUT = 30

# Maximum time between consecutive drinks to be considered in the same 'session'
# (see UserDrinkingSession table)
DRINK_SESSION_TIME_MINUTES = 90

# Maximum time between different UserDrinkingSession records to be considered
# 'concurrent' (see DrinkingSessionGroup table)
GROUP_SESSION_TIME_MINUTES = 90

# The maximum difference between consecutive meter readings that is considered
# valid.
MAX_METER_READING_DELTA = 2200*2

# Minimum and maximum thermo sensor readings (degrees C).
THERMO_SENSOR_RANGE = (-20.0, 80.0)

# Address the kegnet server should bind to.
KB_CORE_DEFAULT_ADDR = 'localhost:9805'

DEFAULT_NEW_USER_WEIGHT = 140
DEFAULT_NEW_USER_GENDER = 'male'

# String name for all taps
ALIAS_ALL_TAPS = '__all_taps__'

# Device names
AUTH_MODULE_CORE_ONEWIRE = 'core.onewire'
AUTH_MODULE_CONTRIB_PHIDGET_RFID = 'contrib.phidget.rfid'

# Maximum time in seconds an authentication token may be missing before the flow
# will be ended.  A value of 0 will cause the fow to be ended immediately when
# the token disappears.
# TODO(mikey): expose configuration in db.
AUTH_TOKEN_MAX_IDLE = {
  AUTH_MODULE_CORE_ONEWIRE: 1,
  AUTH_MODULE_CONTRIB_PHIDGET_RFID: 240,
}

# How often to record a thermo reading?
THERMO_RECORD_DELTA_SECONDS = 60

### Addresses and ports

# Default port for JSON server
JSON_SERVER_PORT_DEFAULT = 9100

### Enums

KB_DEVICE_TYPE = util.Enum(*(
  'DEVICE_TYPE_UNKNOWN',
  'DEVICE_TYPE_FLOW_DEV',
  'DEVICE_TYPE_RELAY',
  'DEVICE_TYPE_THERMO',
))

### Exceptions

class ConfigurationError(Exception):
  """Raised when the system is misconfigured"""



# Copyright 2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Simple event-passing mechanisms.

This module implements a very simple inter-process event passing system
(EventHub), and corresponding message class (Event).
"""

import logging
import Queue

class EventHub(object):
  """Central sink and publish of events."""
  def __init__(self):
    self._event_listeners = set()
    self._event_queue = Queue.Queue()
    self._logger = logging.getLogger('eventhub')

  def AddListener(self, listener):
    """Attach a listener, to be notified on receipt of a new event.

    The listener must implement the PostEvent(event) method.
    """
    if listener not in self._event_listeners:
      self._event_listeners.add(listener)

  def RemoveListener(self, listener):
    """Remove (by reference) an already-listening listener."""
    if listener in self._event_listeners:
      self._event_listeners.remove(listener)

  def PublishEvent(self, event):
    """Add a new event to the queue of events to publish.

    Events are dispatched to listeners in the DispatchNextEvent method.
    """
    self._event_queue.put(event)

  def _IterEventListeners(self):
    """Iterate through all listeners."""
    for listener in self._event_listeners:
      yield listener

  def _WaitForEvent(self, timeout=None):
    """Wait for a new event to be enqueued."""
    try:
      ev = self._event_queue.get(block=True, timeout=timeout)
    except Queue.Empty:
      ev = None
    return ev

  def DispatchNextEvent(self, timeout=None):
    """Wait for an event, and dispatch it to all listeners."""
    ev = self._WaitForEvent(timeout)
    if ev:
      self._logger.debug('Publishing event: %s %s' % (ev.DESCRIPTOR.name, ev))
      for listener in self._IterEventListeners():
        listener.PostEvent(ev)




#!/usr/bin/env python

"""Unittest for kegboard module"""

import os
import unittest
import struct

from pykeg.hw.kegboard import kegboard

TESTDATA_PATH = os.path.join(os.path.dirname(kegboard.__file__), 'testdata')
CAP_FILE = os.path.join(TESTDATA_PATH, 'one_flow_active.bin')


class MessageTestCase(unittest.TestCase):
  def testMessageCreate(self):
    hello_bytes = kegboard.KBSP_PREFIX + '\x01\x00\x04\x00\x01\x02\x03\x00\x3f\x29\r\n'
    m = kegboard.HelloMessage(bytes=hello_bytes)
    self.assertEqual(m.firmware_version, 3)

    m = kegboard.GetMessageForBytes(hello_bytes)
    self.assertEqual(m.firmware_version, 3)
    print m


class KegboardReaderTestCase(unittest.TestCase):
  def testBasicUse(self):
    fd = open(CAP_FILE, 'rb')
    kbr = kegboard.KegboardReader(fd)

    # read the first 8 messages
    messages = []
    for i in xrange(8):
      messages.append(kbr.GetNextMessage())
    print 'messages:'
    print '\n'.join('  %s' % msg for msg in messages)

    hello_message = kegboard.HelloMessage()
    hello_message.SetValue('firmware_version', 3)
    self.assertEqual(messages[0], hello_message)

    onewire_message = kegboard.OnewirePresenceMessage()
    onewire_message.SetValue('device_id', 0)
    self.assertEqual(messages[1], onewire_message)

    message_bytes = messages[2].ToBytes()
    new_message = kegboard.GetMessageForBytes(message_bytes)
    self.assertEqual(message_bytes, new_message.ToBytes())

  def testAgainstBogusData(self):
    pass

if __name__ == '__main__':
  import logging
  logging.basicConfig()
  unittest.main()



from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

import datetime

from pykeg.core import models


class Command(BaseCommand):
  help = u'Compress temperature records'
  args = '<none>'

  def handle(self, *args, **options):
    if len(args) != 0:
      raise CommandError('No arguments required')

    models.Thermolog.CompressLogs()



# Copyright 2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Utility module for summing flows."""

import datetime
import logging

class FlowMeter(object):
  """Represents a path of fluid volume."""
  def __init__(self, name, max_delta=0, start_ticks=None):
    self._name = name
    self._max_delta = max_delta
    self._last_ticks = start_ticks
    self._total_ticks = 0
    self._last_activity = datetime.datetime.fromtimestamp(0)
    self._logger = logging.getLogger('flowmeter-%s' % self._name)

  def __str__(self):
    return "<FlowMeter %s ticks=%i>" % (self._name, self.GetTicks())

  def SetTicks(self, ticks, when=None):
    """Report the instantaneous reading of the meter.

    Clients of this interface should provide a monotonically-increasing,
    "odometer-style" value as the ticks parameter.  Internally, the FlowMeter
    instance will compute the difference in ticks based on any previously
    reported ticks.

    If a value for ticks has not yet been reported, the value is simply saved
    for the next report (without incrementing the total ticks.)
    """
    ticks = long(ticks)
    if ticks < 0:
      raise ValueError, "SetTicks only accepts positive values"

    last_reading = self._last_ticks
    if last_reading is None:
      last_reading = ticks

    self._last_ticks = ticks

    if ticks >= last_reading:
      # Normal case: zero or positive delta from last reading.
      delta = ticks - last_reading
    else:
      # Overflow case: ticks < last ticks
      # If the value is less than the max delta, then use it. Some ticks will be
      # lost, since we can't be sure where the overflow occurred. Drop the
      # reading otherwise.
      self._logger.warning('New ticks report less than previous: %i < %i' %
          (ticks, last_reading))
      if ticks < self._max_delta:
        # The value is less than the max delta. In this case, assume the device
        # wrapped around or reset to zero, and use the current reading as the
        # absolute value. Some ticks may be lost.
        # TODO(mikey): should warn and/or accumulate stats
        delta = ticks
      else:
        # The value is greater than the maximum delta; drop the reading, since
        # we cannot safely accept it.
        # TODO(mikey): should warn and/or accumulate stats
        delta = 0

    if self._max_delta and delta > self._max_delta:
      self._logger.warning('Delta greater than maximum, dropping reading. '
          '(%i > %i)' % (delta, self._max_delta))
      # TODO(mikey): accumulate stat somewhere
      # TODO(mikey): add sequence number to updates, or guard against reset by
      # some other means..?
      return 0

    # If there was actually a change, increment total ticks and update activity
    # time.
    if delta > 0:
      self._total_ticks += delta
      self._SetActivity(when)
      return delta

    # No activity
    return 0

  def _SetActivity(self, when=None):
    """Internal method to update the last activity timestamp."""
    if not when:
      when = datetime.datetime.now()
    self._last_activity = when

  ### Accessors
  def GetTicks(self):
    return self._total_ticks

  def GetLastReading(self):
    return self._last_ticks

  def GetName(self):
    return self._name

  def GetLastActivity(self):
    return self._last_activity

  def GetIdleTime(self, now=None):
    if not now:
      now = datetime.datetime.now()
    return now - self._last_activity




# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous protobuf-related utils."""

import sys

if sys.version_info[:2] < (2, 6):
  import simplejson as json
else:
  import json

def ProtoMessageToDict(message):
  ret = {}
  if not message.IsInitialized():
    raise ValueError, 'Message not initialized'
  for descriptor, value in message.ListFields():
    if descriptor.type == descriptor.TYPE_MESSAGE:
      ret[descriptor.name] = ProtoMessageToDict(value)
    else:
      ret[descriptor.name] = value
  return ret

def DictToProtoMessage(values, out_message):
  for name, field in out_message.DESCRIPTOR.fields_by_name.iteritems():
    if name not in values:
      if field.label == field.LABEL_REQUIRED:
        raise ValueError, "Missing required field %s" % name
      continue

    value = values.get(name)
    if field.type == field.TYPE_MESSAGE:
      inner_message = getattr(out_message, name)
      value = DictToProtoMessage(value, inner_message)
    else:
      setattr(out_message, name, value)
  return out_message




# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.db import connection
from django.core.management.base import CommandError
from django.core.management.base import NoArgsCommand

from pykeg.core import models
from pykeg.core.management.commands.common import progbar

class Command(NoArgsCommand):
  help = u'Regenerate all cached stats.'
  args = '<none>'

  def handle(self, **options):
    drinks = models.Drink.objects.all()

    pos = 0
    count = drinks.count()
    for d in drinks:
      pos += 1
      progbar('clear drink sessions', pos, count)
      d.session = None
      d.save()
    print ''

    print 'deleting old sessions..',
    try:
      cursor = connection.cursor()
      for table in ('core_drinkingsession', 'core_kegsessionchunk',
        'core_usersessionchunk', 'core_sessionchunk'):
        cursor.execute('TRUNCATE TABLE `%s`' % table)
      print 'truncate successful'
    except Exception, e:
      models.DrinkingSession.objects.all().delete()
      print 'orm delete successful'

    pos = 0
    count = drinks.count()
    for d in drinks.order_by('endtime'):
      pos += 1
      progbar('calc new sessions', pos, count)
      sess = models.DrinkingSession.AssignSessionForDrink(d)
    print ''

    print 'done!'



# Copyright 2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

import random

from django.db import models
from django.contrib.auth.models import User

from pykeg.core import models as core_models
from pykeg.core import units

class UserTwitterLink(models.Model):
  """Maps a user to a particular twitter account."""
  user_profile = models.OneToOneField(core_models.UserProfile, primary_key=True)
  twitter_name = models.CharField(max_length=256)
  validated = models.BooleanField(default=True)


class TweetLog(models.Model):
  twitter_id = models.CharField(max_length=256)
  tweet = models.CharField(max_length=256)


class DrinkTweetLog(models.Model):
  tweet_log = models.ForeignKey(TweetLog)
  drink = models.ForeignKey(core_models.Drink)


class DrinkClassification(models.Model):
  name = models.CharField(max_length=256)
  minimum_volume_ml = models.FloatField()

  def __str__(self):
    vol = units.Quantity(self.minimum_volume_ml, units.UNITS.Milliliter)
    oz = vol.ConvertTo.Ounce
    return '%s (%.1foz or more)' % (self.name, oz)

  @classmethod
  def GetClassForDrink(cls, drink):
    all_classes = list(cls.objects.all().order_by('minimum_volume_ml'))
    if not all_classes:
      return None

    best = all_classes[0]
    if drink.volume_ml < best.minimum_volume_ml:
      return None
    for drink_class in all_classes[1:]:
      if drink.volume_ml < drink_class.minimum_volume_ml:
        return best
      best = drink_class
    return best


class DrinkRemark(models.Model):
  drink_class = models.ForeignKey(DrinkClassification)
  remark = models.CharField(max_length=256)

  def  __str__(self):
    return "%s [%s]" % (self.remark, self.drink_class.name)

  @classmethod
  def GetRemarkForDrink(cls, drink):
    drink_class = DrinkClassification.GetClassForDrink(drink)
    if not drink_class:
      return None
    remarks = list(drink_class.drinkremark_set.all())
    if not remarks:
      return None
    return random.choice(remarks)



#!/usr/bin/env python

"""Unittest for util module"""

import unittest
import util

class EnumTestCase(unittest.TestCase):
  def setUp(self):
    self.enum = util.Enum(*(
      'Foo',
      ('Bar', 3),
      'Dip',
      ('Baz', 10),
    ))

  def testBasicUse(self):
    self.assertNotEqual(self.enum.Foo, self.enum.Bar)
    print str(self.enum)
    self.assertEqual(self.enum.Foo.Value, 0)
    self.assertEqual(self.enum.Bar.Value, 3)
    self.assertEqual(self.enum.Dip.Value, 4)
    self.assertEqual(self.enum.Baz.Value, 10)
    self.assertEqual(str(self.enum.Bar), 'Bar')


class SimpleGraphTestCase(unittest.TestCase):
  def setUp(self):
    self.graph = util.SimpleGraph((
        ('a', 'b'),
        ('a', 'c'),
        ('c', 'd'),
        ('d', 'e'),
        ('c', 'e'),
        ('e', 'f'),
    ))

  def testReverseMappings(self):
    self.assertEquals(['a', 'c'], self.graph.ShortestPath('a', 'c'))
    self.assertEquals(['a', 'c', 'e', 'f'], self.graph.ShortestPath('a', 'f'))

if __name__ == '__main__':
  unittest.main()



from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from pykeg.core import defaults

class Command(BaseCommand):
  help = u'Set defaults in a new kegbot database.'
  args = '<none>'

  def handle(self, *args, **options):
    if len(args) != 0:
      raise CommandError('No arguments required')

    if defaults.db_is_installed():
      print 'Error: database already installed, defaults cannot be set'
      return

    defaults.set_defaults()



from django import forms

from pykeg.core import models
from pykeg.beerdb import models as bdb

ALL_TAPS = models.KegTap.objects.all()
ALL_SIZES = models.KegSize.objects.all()
ALL_BEER_TYPES = bdb.BeerType.objects.all().order_by('name')

class GeneralSettingsForm(forms.Form):
  name = forms.CharField(help_text='Name of this Kegbot system')

class ChangeKegForm(forms.Form):
  tap = forms.ModelChoiceField(queryset=ALL_TAPS,
      empty_label=None,
      help_text='Select tap to add/replace keg')

  keg_size = forms.ModelChoiceField(queryset=ALL_SIZES,
      empty_label=None,
      help_text='Size of the new keg')

  beer_type = forms.ModelChoiceField(queryset=ALL_BEER_TYPES,
      empty_label=None,
      help_text='Choose existing type')

  cost = forms.FloatField(required=False, min_value=0,
      help_text='Price paid (optional)')
  description = forms.CharField(required=False,
      help_text='Public description of this specific keg (optional)')

class TapForm(forms.ModelForm):
  class Meta:
    model = models.KegTap

#BeerTypeFormSet = inlineformset_factory(models.Brewer, models.BeerType)



# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb API client."""

import datetime
import sys
import types

from pykeg import settings
from pykeg.core import models_pb2
from pykeg.core.protoutil import DictToProtoMessage
from pykeg.web.api import common

try:
  from urllib.parse import urlencode
  from urllib.request import urlopen
  from urllib.error import URLError
except ImportError:
  from urllib import urlencode
  from urllib2 import urlopen
  from urllib2 import URLError

try:
  import json
except ImportError:
  try:
    import simplejson as json
  except ImportError:
    import django.utils.simplejson as json

import gflags
FLAGS = gflags.FLAGS

_DEFAULT_URL = 'http://localhost:8000/api/'
if hasattr(settings, 'KEGWEB_BASE_URL'):
  _DEFAULT_URL = '%s/api' % getattr(settings, 'KEGWEB_BASE_URL')

gflags.DEFINE_string('krest_url', _DEFAULT_URL,
    'Base URL for the Kegweb HTTP api.')

class ApiError(Exception):
  """Raised when the remote side returns an error."""

class ServerError(ApiError):
  """Raised when a malformed response is received from the server."""

class RemoteError(ApiError):
  """Raised when the remote API raises an error."""
  def __init__(self, code, message=None):
    self.code = code
    self.message = message


class KrestClient:
  """Kegweb RESTful API client."""
  def __init__(self, base_url=None, api_auth_token=None):
    if not base_url:
      base_url = FLAGS.krest_url
    self._base_url = base_url
    self._api_auth_token = api_auth_token

  def _Encode(self, s):
    return unicode(s).encode('utf-8')

  def _EncodePostData(self, post_data):
    if not post_data:
      return None
    return urlencode(dict(((k, self._Encode(v)) for k, v in
        post_data.iteritems() if v is not None)))

  def _GetURL(self, endpoint, params=None):
    param_str = ''
    if params:
      param_str = '?%s' % urlencode(params)

    base = self._base_url.rstrip('/')

    # Strip both leading and trailing slash from endpoint: single leading and
    # trailing slashes will be added by the string formatter.  (The trailing
    # slash is required for POSTs to Django.)
    endpoint = endpoint.strip('/')
    return '%s/%s/%s' % (base, endpoint, param_str)

  def SetAuthToken(self, api_auth_token):
    self._api_auth_token = api_auth_token

  def DoGET(self, endpoint, params=None):
    """Issues a GET request to the endpoint, and retuns the result.

    Keyword arguments are passed to the endpoint as GET arguments.

    For normal responses, the return value is the Python JSON-decoded 'result'
    field of the response.  If the response is an error, a RemoteError exception
    is raised.

    If there was an error contacting the server, or in parsing its response, a
    ServerError is raised.
    """
    return self._FetchResponse(endpoint, params=params)

  def DoPOST(self, endpoint, post_data, params=None):
    """Issues a POST request to the endpoint, and returns the result.

    For normal responses, the return value is the Python JSON-decoded 'result'
    field of the response.  If the response is an error, a RemoteError exception
    is raised.

    If there was an error contacting the server, or in parsing its response, a
    ServerError is raised.
    """
    return self._FetchResponse(endpoint, params=params, post_data=post_data)

  def _FetchResponse(self, endpoint, params=None, post_data=None):
    """Issues a POST or GET request, depending on the arguments."""
    if params is None:
      params = {}
    else:
      params = params.copy()

    # Request a full response, unless caller specified otherwise.
    if params and 'full' in params:
      if params['full'] != 0:
        params['full'] = 1
    else:
      params['full'] = 1

    # If we have an api token, attach it.  Prefer to attach it to POST data, but
    # use GET if there is no POST data.
    if self._api_auth_token:
      if post_data:
        post_data['api_auth_token'] = self._api_auth_token
      else:
        params['api_auth_token'] = self._api_auth_token

    url = self._GetURL(endpoint, params=params)
    encoded_post_data = self._EncodePostData(post_data)

    try:
      # Issue a GET or POST (urlopen will decide based on encoded_post_data).
      response_data = urlopen(url, encoded_post_data).read()
    except URLError, e:
      raise ServerError('Caused by: %s' % e)

    return self._DecodeResponse(response_data)

  def _DecodeResponse(self, response_data):
    """Decodes the string `response_data` as a JSON response.

    For normal responses, the return value is the Python JSON-decoded 'result'
    field of the response.  If the response is an error, a RemoteError exception
    is raised.
    """
    # Decode JSON.
    try:
      d = json.loads(response_data)
    except ValueError, e:
      raise ServerError('Malformed response: %s' % e)

    if 'error' in d:
      # Response had an error: translate to exception.
      err = d.get('error')
      if type(err) != types.DictType:
        raise ValueError('Invalid error response from server')
      code = err.get('code', 'Error')
      message = err.get('message', None)
      e = common.ErrorCodeToException(code, message)
      raise e
    elif 'result' in d:
      # Response was OK, return the result.
      return d.get('result')
    else:
      # WTF?
      raise ValueError('Invalid response from server: missing result or error')

  def _DictListToProtoList(self, d_list, proto_obj):
    return [DictToProtoMessage(m, proto_obj) for m in d_list]

  def RecordDrink(self, tap_name, ticks, volume_ml=None, username=None,
      pour_time=None, duration=None, auth_token=None):
    endpoint = '/tap/%s' % tap_name
    post_data = {
      'tap_name' : tap_name,
      'ticks' : ticks,
      'volume_ml' : volume_ml,
      'username' : username,
      'auth_token' : auth_token,
    }
    if pour_time:
      post_data['pour_time'] = pour_time
      post_data['now'] = datetime.datetime.now()
    return self.DoPOST(endpoint, post_data=post_data)

  def LogSensorReading(self, sensor_name, temperature, when=None):
    endpoint = '/thermo-sensor/%s' % (sensor_name,)
    post_data = {
      'temp_c': float(temperature),
    }
    # TODO(mikey): include post data
    res = self.DoPOST(endpoint, post_data=post_data)
    ret = DictToProtoMessage(res, models_pb2.ThermoLog)
    return ret

  def TapStatus(self, full=True):
    """Gets the status of all taps."""
    params = {'full' : full}
    response = self.DoGET('tap', params)
    return self._DictListToProtoList(response, models_pb2.KegTap())

  def GetToken(self, auth_device, token_value, full=True):
    url = 'auth-token/%s.%s' % (auth_device, token_value)
    params = {'full' : full}
    response = self.DoGET(url, params)
    return DictToProtoMessage(response, models_pb2.AuthenticationToken())

  def LastDrinks(self, full=True):
    """Gets a list of the most recent drinks."""
    params = {'full' : full}
    response = self.DoGET('last-drinks', params)
    return self._DictListToProtoList(response, models_pb2.Drink())

  def AllDrinks(self, full=True):
    """Gets a list of all drinks."""
    params = {'full' : full}
    response = self.DoGET('drink', params)
    return self._DictListToProtoList(response, models_pb2.Drink())


def main():
  c = KrestClient()

  print '== record a drink =='
  print c.RecordDrink('kegboard.flow0', 2200)
  print ''

  print '== tap status =='
  for t in c.TapStatus():
    print t
    print ''

  print '== last drinks =='
  for d in c.LastDrinks():
    print d
    print ''

if __name__ == '__main__':
  FLAGS(sys.argv)
  main()



from django.conf.urls.defaults import *
from pykeg.core import features

urlpatterns = patterns('pykeg.web.account.views',
    url(r'^$', 'account_main', name='kb-account-main'),
    url(r'^mugshot/$', 'edit_mugshot', name='account-mugshot'),
)

if features.use_facebook():
  urlpatterns += patterns('',
      url(r'fb-settings/$', 'pykeg.web.contrib.facebook.views.account_settings', name='fb-account-settings'),
  )

if features.use_twitter():
  urlpatterns += patterns('',
      url(r'twitter-settings/$', 'pykeg.web.contrib.twitter.views.account_settings', name='twitter-account-settings'),
  )



from django.conf import settings

"""Methods to query configuration of optional features."""

def use_facebook():
  return 'socialregistration' in settings.INSTALLED_APPS and \
      'socialregistration.middleware.FacebookMiddleware' in settings.MIDDLEWARE_CLASSES and \
      'socialregistration.auth.FacebookAuth' in settings.AUTHENTICATION_BACKENDS and \
      getattr(settings, 'FACEBOOK_API_KEY', None) is not None and \
      getattr(settings, 'FACEBOOK_SECRET_KEY', None) is not None

def use_twitter():
  return 'pykeg.contrib.twitter' in settings.INSTALLED_APPS



# Copyright 2008 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Various unit conversion routines."""

import types

from pykeg.core import util

UNITS = util.Enum(*(
  'Liter',
  'Milliliter',
  'Microliter',
  'Ounce',
  'Pint',
  'USGallon',
  'ImperialGallon',
  'TwelveOunceBeer',
  'HalfBarrelKeg',
  'KbMeterTick', # aka Vision2000 (2200 pulses/liter)
  'PonyKeg',
  'Cup',
  'Quart',
  'Hogshead',
))

# This constant defines the unit type used for the database "volume" field
RECORD_UNIT = UNITS.Milliliter

CONVERSIONS = {
    (UNITS.Milliliter, UNITS.KbMeterTick) : 2.2,
    (UNITS.Liter, UNITS.Milliliter) : 1e3,
    (UNITS.Liter, UNITS.Microliter) : 1e6,
    (UNITS.Liter, UNITS.Ounce) : 33.8140227,
    (UNITS.USGallon, UNITS.Ounce) : 128.0,
    (UNITS.Cup, UNITS.Ounce) : 8.0,
    (UNITS.HalfBarrelKeg, UNITS.USGallon) : 15.5,
    (UNITS.HalfBarrelKeg, UNITS.PonyKeg) : 2.0,
    (UNITS.Pint, UNITS.Ounce) : 16.0,
    (UNITS.TwelveOunceBeer, UNITS.Ounce) : 12.0,
    (UNITS.Hogshead, UNITS.USGallon) : 63.0,
    (UNITS.Quart, UNITS.Ounce) : 32.0,
    (UNITS.ImperialGallon, UNITS.Liter) : 4.54609,
}

class ConversionError(Exception):
  """Raised if a conversion is not possible"""

class _UnitConverter:
  def __init__(self, table):
    self._table = {}
    self._table.update(table)

    all_units = set()
    for unit_from, unit_to in table.keys():
      all_units.add(unit_from)
      all_units.add(unit_to)

    # Generate reverse conversion factors for those given
    for k, v in table.iteritems():
      unit_from, unit_to = k
      reverse_value = 1.0/v
      reverse_key = (unit_to, unit_from)
      # TODO: Should we assert if any existing value in the table does not match
      # computed value? For now, just skipping.
      if reverse_key not in self._table:
        self._table[reverse_key] = reverse_value

    unit_graph = util.SimpleGraph(self._table.keys())

    # Generate all possible conversion factors
    for unit_from in all_units:
      for unit_to in all_units:
        if unit_from == unit_to:
          continue
        new_key = (unit_from, unit_to)
        if new_key in self._table:
          continue

        # Attempt to find a path to these units
        path = unit_graph.ShortestPath(unit_from, unit_to)
        if not path:
          # Not possible, just continue
          continue
        value = 1.0
        prev_unit = path[0]
        for unit in path[1:]:
          value = self.Convert(value, prev_unit, unit)
          prev_unit = unit
        self._table[new_key] = value

  def CanConvert(self, from_unit, to_unit):
    return (from_unit, to_unit) in self._table

  def Convert(self, amt, from_unit, to_unit):
    if from_unit == to_unit:
      return amt
    k = (from_unit, to_unit)
    if k not in self._table:
      raise ConversionError, "Don't know how to convert %s to %s" % (from_unit,
                                                                     to_unit)
    return float(amt) * self._table[k]

UnitConverter = _UnitConverter(CONVERSIONS)


class Quantity:
  DEFAULT_UNITS = UNITS.Milliliter

  class _AttributeConverter:
    def __init__(self, qty=0):
      self._qty = qty

    def __getattr__(self, name):
      qty_obj = self.__dict__['_qty']
      return UnitConverter.Convert(qty_obj.Amount(),
                                   qty_obj.units(),
                                   getattr(UNITS, name))

  def __init__(self, amount, units=None, from_units=None):
    if units is None:
      units = self.DEFAULT_UNITS
    self._units = units

    if from_units is None:
      from_units = self._units

    self._amount = UnitConverter.Convert(amount, from_units, self.units())
    self.ConvertTo = Quantity._AttributeConverter(self)

  def __add__(self, other, subtract=False):
    val = 0
    if isinstance(other, (types.IntType, types.LongType, types.FloatType)):
      val += other
    elif isinstance(other, Quantity):
      val += other.Amount(self.units())
    else:
      raise TypeError
    if subtract:
      val = self.Amount() - val
    else:
      val += self.Amount()
    return Quantity(val, self.units())

  def __sub__(self, other):
    return self.__add__(other, subtract=True)

  def __cmp__(self, other):
    if isinstance(other, (types.IntType, types.LongType, types.FloatType)):
      return cmp(self.Amount(), other)
    elif isinstance(other, Quantity):
      return cmp(self.Amount(), other.Amount(in_units=self.units()))
    else:
      raise TypeError

  # TODO: should have just subclassed float
  def __int__(self):
    return int(self.Amount())

  def __long__(self):
    return long(self.Amount())

  def __float__(self):
    return float(self.Amount())

  def units(self):
    return self._units

  def IncAmount(self, amt, units=None):
    if units is None:
      self._amount += amt
    else:
      self._amount += UnitConverter.Convert(amt, units, self.units())

  def SetAmount(self, amt, units=None):
    self._amount= 0
    self.IncAmount(amt, units)

  def Clear(self):
    self.SetAmount(0)

  def Amount(self, in_units=None):
    # Same as self.ConvertTo.<in_units>
    if in_units is None:
      in_units = self.units()
    return UnitConverter.Convert(self._amount, self.units(), in_units)



#!/usr/bin/env python

"""Unittest for kegnet module"""

import asyncore
import logging
import unittest

import gflags

from pykeg.core.net import kegnet

FLAGS = gflags.FLAGS

LOGGER = logging.getLogger('myunittest')

class KegnetTestCase(unittest.TestCase):
  def setUp(self):
    self.server = kegnet.KegnetServer(name='kegnet', kb_env=None,
        addr=FLAGS.kb_core_bind_addr)

  def tearDown(self):
    pass

  def testSimpleFlow(self):
    print 'Starting server'
    self.server.StartServer()
    print 'Done start'
    print 'Looping'
    asyncore.loop(timeout=1, count=1)
    print 'New client'
    client = kegnet.KegnetClient()
    client.SendFlowStart('mytap')
    asyncore.loop(timeout=1, count=1)
    print 'Done'

if __name__ == '__main__':
  import sys
  logging.basicConfig(stream=sys.stderr)
  LOGGER.setLevel(logging.DEBUG)
  LOGGER.info('here i am!')
  unittest.main()



# -*- coding: latin-1 -*-
# Copyright 2009 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import admin

from pykeg.core import models
from pykeg.core import util

admin.site.register(models.UserPicture)
admin.site.register(models.UserProfile)

admin.site.register(models.KegSize)
admin.site.register(models.KegbotSite)

class KegTapAdmin(admin.ModelAdmin):
  list_display = ('name', 'meter_name', 'current_keg')
admin.site.register(models.KegTap, KegTapAdmin)

class KegAdmin(admin.ModelAdmin):
  list_display = ('id', 'type')
admin.site.register(models.Keg, KegAdmin)

class DrinkAdmin(admin.ModelAdmin):
  list_display = ('id', 'user', 'keg', 'endtime')
  list_filter = ('keg', 'status', 'endtime')
  search_fields = ('id', 'user__username')
admin.site.register(models.Drink, DrinkAdmin)

class AuthenticationTokenAdmin(admin.ModelAdmin):
  list_display = ('auth_device', 'user', 'token_value', 'enabled', 'IsActive')
  list_filter = ('auth_device', 'enabled')
  search_fields = ('user__username', 'token_value')
admin.site.register(models.AuthenticationToken, AuthenticationTokenAdmin)

admin.site.register(models.BAC)
admin.site.register(models.DrinkingSession)

class ThermoSensorAdmin(admin.ModelAdmin):
  list_display = ('nice_name', 'raw_name')

admin.site.register(models.ThermoSensor, ThermoSensorAdmin)

def thermolog_deg_c(obj):
  return '%.2f C' % (obj.temp,)

def thermolog_deg_f(obj):
  return '%.2f F' % (util.CtoF(obj.temp),)

class ThermologAdmin(admin.ModelAdmin):
  list_display = ('sensor', thermolog_deg_c, thermolog_deg_f, 'time')
  list_filter = ('sensor', 'time')

admin.site.register(models.Thermolog, ThermologAdmin)

class ThermoSummaryLogAdmin(admin.ModelAdmin):
  list_display = ('sensor', 'date', 'min_temp', 'max_temp', 'mean_temp')
  list_filter = ('sensor', 'date')

admin.site.register(models.ThermoSummaryLog, ThermoSummaryLogAdmin)

admin.site.register(models.RelayLog)

class ConfigAdmin(admin.ModelAdmin):
  list_display = ('key', 'value')
  search_fields = ('key', 'value')
admin.site.register(models.Config, ConfigAdmin)



#!/usr/bin/env python
#
# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegbot LCD daemon."""

from pykeg.core import importhacks

import datetime
import gflags
import Queue
import sys
import time

from pykeg.core import kb_app
from pykeg.core import units
from pykeg.core import util
from pykeg.core.net import kegnet
from pykeg.core.net import kegnet_pb2

from pykeg.web.api.krest import KrestClient

try:
  import lcdui
except ImportError:
  print>>sys.stderr, "Error: lcdui could not be imported."
  print>>sys.stderr, "(Try: sudo easy_install --upgrade pylcdui)"
  sys.exit(1)

from lcdui.ui import frame
from lcdui.ui import ui
from lcdui.ui import widget

DEVICE_CFA_635 = 'cfa635'
DEVICE_MTXORB = 'mtxorb20x4'

DEVICE_CHOICES = ', '.join((
  DEVICE_CFA_635,
  DEVICE_MTXORB
))

FLAGS = gflags.FLAGS

gflags.DEFINE_string('lcd_device_path', '/dev/ttyUSB0',
    'Path to lcd device file.')

gflags.DEFINE_integer('lcd_baud_rate', 115200,
    'LCD baud rate.')

gflags.DEFINE_integer('backlight_timeout', 300,
    'Number of seconds of inactivity before LCD backlight will be disabled.')

gflags.DEFINE_string('lcd_device_type', DEVICE_CFA_635,
    'Type of LCD device. Choices are: %s' % DEVICE_CHOICES)

gflags.DEFINE_integer('rotation_time', 10,
    'Seconds to pause between rotated tap/drink info screens.',
    lower_bound=1)

gflags.DEFINE_integer('krest_update_interval', 60,
    'Time between periodic refreshes of tap and drink information '
    'by the Kegweb REST client.', lower_bound=10)


class KegUi:
  STATE_MAIN = 'main'
  STATE_POUR = 'pour'

  def __init__(self, path=None):
    if not path:
      path = FLAGS.lcd_device_path

    if FLAGS.lcd_device_type == DEVICE_CFA_635:
      from lcdui.devices import CrystalFontz
      self._lcdobj = CrystalFontz.CFA635Display(path, FLAGS.lcd_baud_rate)
    elif FLAGS.lcd_device_type == DEVICE_MTXORB:
      from lcdui.devices import MatrixOrbital
      self._lcdobj = MatrixOrbital.MatrixOrbitalDisplay(path,
          FLAGS.lcd_baud_rate)
    else:
      raise ValueError, "Bad device type: %s" % FLAGS.lcd_device_type

    self._lcdui = ui.LcdUi(self._lcdobj, FLAGS.backlight_timeout)
    self._last_flow_status = None

    self._multiframe = self._lcdui.FrameFactory(frame.MultiFrame)
    self._splash_frame = self._BuildSplashFrame()
    self._pour_frame = self._BuildPourFrame()
    self._last_drink_frame = self._BuildLastDrinkFrame()

    self._multiframe.AddFrame(self._splash_frame, 5)

    self._flow_status_by_tap = {}

    self._state = None
    self._SetState(self.STATE_MAIN)

  def _BuildSplashFrame(self):
    f = self._lcdui.FrameFactory(frame.TextFrame)
    f.AddLine(' ================== ')
    f.AddLine('       kegbot!      ')
    f.AddLine('      beer you.     ')
    f.AddLine(' ================== ')
    return f

  def _BuildPourFrame(self):
    f = self._lcdui.FrameFactory(frame.Frame)
    f.BuildWidget(widget.LineWidget, name='line0',
        contents='_now pouring________', row=0, col=0)
    f.BuildWidget(widget.LineWidget, name='user',
        prefix='|user: ', row=1, col=0)
    f.BuildWidget(widget.LineWidget, name='ounces',
        prefix='|oz: ', row=2, col=0)
    f.BuildWidget(widget.LineWidget, name='line3',
        contents='|                   ', row=3, col=0)
    return f

  def _BuildLastDrinkFrame(self):
    f = self._lcdui.FrameFactory(frame.Frame)
    f.BuildWidget(widget.LineWidget, name='line0',
        contents='_last pour__________', row=0, col=0)
    f.BuildWidget(widget.LineWidget, name='user',
        prefix='user: ', row=1, col=0)
    f.BuildWidget(widget.LineWidget, name='when',
        prefix='when: ', row=2, col=0)
    f.BuildWidget(widget.LineWidget, name='size',
        prefix='size: ', row=3, col=0)
    return f

  def _UpdateRotation(self, tap_status_list):
    """Builds a ui.MultiFrame instance from tap status."""
    self._multiframe.frames().clear()
    self._multiframe.AddFrame(self._splash_frame, 5.0)
    self._multiframe.AddFrame(self._last_drink_frame, FLAGS.rotation_time)

    for tap in tap_status_list:
      tap_name = tap.name
      beer_name = 'Unknown Beer'
      brewer_name = 'Unknown Brewer'
      pct_full = 0
      curr_temp = None
      if tap.current_keg:
        pct_full = tap.current_keg.percent_full
        if tap.current_keg.type:
          beer_name = tap.current_keg.type.name
          brewer_name = tap.current_keg.type.brewer.name
      if tap.last_temperature:
        curr_temp = tap.last_temperature.temperature_c

      f = self._BuildTapFrame(tap_name, beer_name, brewer_name, pct_full, curr_temp)
      self._multiframe.AddFrame(f, 10.0)

  def _BuildTapFrame(self, tap_name, beer_name, brewer_name, pct_full,
      curr_temp=None):
    f = self._lcdui.FrameFactory(frame.TextFrame)
    f.SetTitle(str(tap_name))  # note: squash unicode
    f.AddLine(beer_name)
    f.AddLine(brewer_name)
    status = '%.1f%% full' % (100 * pct_full)
    if curr_temp is not None:
      status = '%s  %.1fc' % (status, curr_temp)
    f.AddLine(status)
    return f

  def _UpdateFromFlow(self, flow_update):
    f = self._pour_frame
    user_widget = f.GetWidget('user')
    user_widget.set_contents(str(flow_update.username))

    amt = units.Quantity(flow_update.volume_ml, units.UNITS.Milliliter)
    ounces = amt.ConvertTo.Ounce

    ounces_widget = f.GetWidget('ounces')
    ounces_widget.set_contents('%.2f' % ounces)

  def UpdateFromTapStatus(self, tap_status_list):
    self._UpdateRotation(tap_status_list)

  def UpdateLastDrink(self, username, volume_ml, date):
    vol = units.Quantity(volume_ml, units.UNITS.Milliliter)
    size = '%.2f ounces' % vol.ConvertTo.Ounce
    when = date.strftime('%b %d %I:%M%p').lower()

    f = self._last_drink_frame
    f.GetWidget('user').set_contents(username)
    f.GetWidget('size').set_contents(size)
    f.GetWidget('when').set_contents(when)

  def HandleFlowStatus(self, event):
    tap_name = event.tap_name
    last_status = self._flow_status_by_tap.get(tap_name)
    self._flow_status_by_tap[tap_name] = event

    self._lcdui.Activity()

    if event.state == event.COMPLETED:
      self.UpdateLastDrink(str(event.username), event.volume_ml,
          datetime.datetime.fromtimestamp(event.last_activity_time))
      self._SetState(self.STATE_MAIN)
    else:
      self._SetState(self.STATE_POUR)
      self._UpdateFromFlow(event)

  def _SetState(self, state):
    if state == self._state:
      return

    self._state = state

    if self._state == self.STATE_MAIN:
      self._lcdui.SetFrame(self._multiframe)
    elif self._state == self.STATE_POUR:
      self._lcdui.SetFrame(self._pour_frame)

  def MainLoop(self):
    return self._lcdui.MainLoop()


class LcdDaemonApp(kb_app.App):
  def __init__(self, name='lcd_daemon'):
    kb_app.App.__init__(self, name)

  def _Setup(self):
    if not FLAGS.krest_url:
      self._logger.error('Error: --krest_url must point to <kegweb>/api')
      self._logger.error('Add KEGWEB_BASE_URL to your common_settings.py.')
      sys.exit(1)
    kb_app.App._Setup(self)
    kb_lcdui = KegUi()
    self._AddAppThread(LcdUiThread('kb-lcdui', kb_lcdui))
    self._AddAppThread(KegnetMonitorThread('kegnet-monitor', kb_lcdui))
    self._AddAppThread(KrestUpdaterThread('krest-updater', kb_lcdui))


class LcdUiThread(util.KegbotThread):
  def __init__(self, name, ui):
    util.KegbotThread.__init__(self, name)
    self._lcdui = ui

  def ThreadMain(self):
    self._logger.info('Starting main loop.')
    self._lcdui.MainLoop()
    self._logger.info('Exited main loop.')


class KrestUpdaterThread(util.KegbotThread):
  def __init__(self, name, ui):
    util.KegbotThread.__init__(self, name)
    self._lcdui = ui
    self._client = KrestClient()

  def ThreadMain(self):
    self._logger.info('Starting main loop.')
    while not self._quit:
      self._logger.info('Fetching krest status')
      try:
        tap_status = self._client.TapStatus()
        self._lcdui.UpdateFromTapStatus(tap_status)
        last_drink = tuple(self._client.LastDrinks())[0]
        username = 'unknown'
        if last_drink.user:
          username = str(last_drink.user.username)
        self._lcdui.UpdateLastDrink(username,
            last_drink.volume_ml,
            datetime.datetime.fromtimestamp(last_drink.pour_time))
      except IOError, e:
        self._logger.warning('Could not connect to kegweb: %s' % e)
      time.sleep(FLAGS.krest_update_interval)
    self._logger.info('Exited main loop.')


class LcdKegnetClient(kegnet.SimpleKegnetClient):
  def __init__(self, kb_lcdui, addr=FLAGS.kb_core_addr):
    kegnet.KegnetClient.__init__(self, addr)
    self._kb_lcdui = kb_lcdui

  def onFlowUpdate(self, event):
    self._kb_lcdui.HandleFlowStatus(event)


class KegnetMonitorThread(util.KegbotThread):
  def __init__(self, name, kb_lcdui):
    util.KegbotThread.__init__(self, name)
    self._kb_lcdui = kb_lcdui
    self._client = LcdKegnetClient(self._kb_lcdui)

  def ThreadMain(self):
    self._logger.info('Starting main loop.')
    while not self._quit:
      self._client.serve_forever()
    self._client.stop()


if __name__ == '__main__':
  LcdDaemonApp.BuildAndRun()



import sys

from django.core.management.base import NoArgsCommand
from django.core.management.base import CommandError
from django.contrib.auth.management.commands import createsuperuser

from south.management.commands import migrate
from south.management.commands import syncdb
from pykeg.core.management.commands import kb_set_defaults

from pykeg.core import backend


class Command(NoArgsCommand):
  help = u'Run basic Kegbot setup.'

  def handle(self, **options):
    cfg = None
    try:
      b = backend.KegbotBackend()
      cfg = b.GetConfig()
    except backend.BackendError, e:
      pass

    if cfg:
      print 'Error: Kegbot appears to be already installed. ',
      print 'Try "%s kb_upgrade" instead?' % (sys.argv[0],)
      sys.exit(1)

    self.InitialSetup()

  def _RunCommand(self, cmd, args=None):
    if args is None:
      args = []
    cmdname = cmd.__module__.split('.')[-1]
    arg_str = ' '.join('%s' % a for a in args)
    print '--- Running command: %s %s' % (cmdname, arg_str)
    cmd.run_from_argv([sys.argv[0], cmdname] + args)

  def InitialSetup(self):
    # syncdb
    self._RunCommand(syncdb.Command(), args=['--all', '--noinput', '-v', '0'])

    # migrate
    self._RunCommand(migrate.Command(), args=['--all', '--fake', '-v', '0'])

    # kb_set_defaults
    self._RunCommand(kb_set_defaults.Command())



#!/usr/bin/env python

"""Unittest for crc16 module."""

import struct
import unittest

from pykeg.hw.kegboard import crc16

class CrcTestCase(unittest.TestCase):
  def testCrc(self):
    data = '\x01\x02\x03\x04'
    crc = crc16.crc16_ccitt(data)
    data += struct.pack('<H', crc)
    self.assertEqual(crc16.crc16_ccitt(data), 0)



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import service
from google.protobuf import service_reflection


_FIELDDESCRIPTORPROTO_TYPE = descriptor.EnumDescriptor(
  name='Type',
  full_name='google.protobuf.FieldDescriptorProto.Type',
  filename='Type',
  values=[
    descriptor.EnumValueDescriptor(
      name='TYPE_DOUBLE', index=0, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_FLOAT', index=1, number=2,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_INT64', index=2, number=3,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_UINT64', index=3, number=4,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_INT32', index=4, number=5,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_FIXED64', index=5, number=6,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_FIXED32', index=6, number=7,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_BOOL', index=7, number=8,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_STRING', index=8, number=9,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_GROUP', index=9, number=10,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_MESSAGE', index=10, number=11,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_BYTES', index=11, number=12,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_UINT32', index=12, number=13,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_ENUM', index=13, number=14,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_SFIXED32', index=14, number=15,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_SFIXED64', index=15, number=16,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_SINT32', index=16, number=17,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='TYPE_SINT64', index=17, number=18,
      options=None,
      type=None),
  ],
  options=None,
)

_FIELDDESCRIPTORPROTO_LABEL = descriptor.EnumDescriptor(
  name='Label',
  full_name='google.protobuf.FieldDescriptorProto.Label',
  filename='Label',
  values=[
    descriptor.EnumValueDescriptor(
      name='LABEL_OPTIONAL', index=0, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='LABEL_REQUIRED', index=1, number=2,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='LABEL_REPEATED', index=2, number=3,
      options=None,
      type=None),
  ],
  options=None,
)

_FILEOPTIONS_OPTIMIZEMODE = descriptor.EnumDescriptor(
  name='OptimizeMode',
  full_name='google.protobuf.FileOptions.OptimizeMode',
  filename='OptimizeMode',
  values=[
    descriptor.EnumValueDescriptor(
      name='SPEED', index=0, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='CODE_SIZE', index=1, number=2,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='LITE_RUNTIME', index=2, number=3,
      options=None,
      type=None),
  ],
  options=None,
)

_FIELDOPTIONS_CTYPE = descriptor.EnumDescriptor(
  name='CType',
  full_name='google.protobuf.FieldOptions.CType',
  filename='CType',
  values=[
    descriptor.EnumValueDescriptor(
      name='STRING', index=0, number=0,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='CORD', index=1, number=1,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='STRING_PIECE', index=2, number=2,
      options=None,
      type=None),
  ],
  options=None,
)


_FILEDESCRIPTORSET = descriptor.Descriptor(
  name='FileDescriptorSet',
  full_name='google.protobuf.FileDescriptorSet',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='file', full_name='google.protobuf.FileDescriptorSet.file', index=0,
      number=1, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_FILEDESCRIPTORPROTO = descriptor.Descriptor(
  name='FileDescriptorProto',
  full_name='google.protobuf.FileDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.FileDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='package', full_name='google.protobuf.FileDescriptorProto.package', index=1,
      number=2, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='dependency', full_name='google.protobuf.FileDescriptorProto.dependency', index=2,
      number=3, type=9, cpp_type=9, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='message_type', full_name='google.protobuf.FileDescriptorProto.message_type', index=3,
      number=4, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='enum_type', full_name='google.protobuf.FileDescriptorProto.enum_type', index=4,
      number=5, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='service', full_name='google.protobuf.FileDescriptorProto.service', index=5,
      number=6, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='extension', full_name='google.protobuf.FileDescriptorProto.extension', index=6,
      number=7, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.FileDescriptorProto.options', index=7,
      number=8, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_DESCRIPTORPROTO_EXTENSIONRANGE = descriptor.Descriptor(
  name='ExtensionRange',
  full_name='google.protobuf.DescriptorProto.ExtensionRange',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='start', full_name='google.protobuf.DescriptorProto.ExtensionRange.start', index=0,
      number=1, type=5, cpp_type=1, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='end', full_name='google.protobuf.DescriptorProto.ExtensionRange.end', index=1,
      number=2, type=5, cpp_type=1, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)

_DESCRIPTORPROTO = descriptor.Descriptor(
  name='DescriptorProto',
  full_name='google.protobuf.DescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.DescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='field', full_name='google.protobuf.DescriptorProto.field', index=1,
      number=2, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='extension', full_name='google.protobuf.DescriptorProto.extension', index=2,
      number=6, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='nested_type', full_name='google.protobuf.DescriptorProto.nested_type', index=3,
      number=3, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='enum_type', full_name='google.protobuf.DescriptorProto.enum_type', index=4,
      number=4, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='extension_range', full_name='google.protobuf.DescriptorProto.extension_range', index=5,
      number=5, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.DescriptorProto.options', index=6,
      number=7, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_FIELDDESCRIPTORPROTO = descriptor.Descriptor(
  name='FieldDescriptorProto',
  full_name='google.protobuf.FieldDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.FieldDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='number', full_name='google.protobuf.FieldDescriptorProto.number', index=1,
      number=3, type=5, cpp_type=1, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='label', full_name='google.protobuf.FieldDescriptorProto.label', index=2,
      number=4, type=14, cpp_type=8, label=1,
      default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='type', full_name='google.protobuf.FieldDescriptorProto.type', index=3,
      number=5, type=14, cpp_type=8, label=1,
      default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='type_name', full_name='google.protobuf.FieldDescriptorProto.type_name', index=4,
      number=6, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='extendee', full_name='google.protobuf.FieldDescriptorProto.extendee', index=5,
      number=2, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='default_value', full_name='google.protobuf.FieldDescriptorProto.default_value', index=6,
      number=7, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.FieldDescriptorProto.options', index=7,
      number=8, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
    _FIELDDESCRIPTORPROTO_TYPE,
    _FIELDDESCRIPTORPROTO_LABEL,
  ],
  options=None)


_ENUMDESCRIPTORPROTO = descriptor.Descriptor(
  name='EnumDescriptorProto',
  full_name='google.protobuf.EnumDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.EnumDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='value', full_name='google.protobuf.EnumDescriptorProto.value', index=1,
      number=2, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.EnumDescriptorProto.options', index=2,
      number=3, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_ENUMVALUEDESCRIPTORPROTO = descriptor.Descriptor(
  name='EnumValueDescriptorProto',
  full_name='google.protobuf.EnumValueDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.EnumValueDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='number', full_name='google.protobuf.EnumValueDescriptorProto.number', index=1,
      number=2, type=5, cpp_type=1, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.EnumValueDescriptorProto.options', index=2,
      number=3, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_SERVICEDESCRIPTORPROTO = descriptor.Descriptor(
  name='ServiceDescriptorProto',
  full_name='google.protobuf.ServiceDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.ServiceDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='method', full_name='google.protobuf.ServiceDescriptorProto.method', index=1,
      number=2, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.ServiceDescriptorProto.options', index=2,
      number=3, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_METHODDESCRIPTORPROTO = descriptor.Descriptor(
  name='MethodDescriptorProto',
  full_name='google.protobuf.MethodDescriptorProto',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.MethodDescriptorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='input_type', full_name='google.protobuf.MethodDescriptorProto.input_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='output_type', full_name='google.protobuf.MethodDescriptorProto.output_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='options', full_name='google.protobuf.MethodDescriptorProto.options', index=3,
      number=4, type=11, cpp_type=10, label=1,
      default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_FILEOPTIONS = descriptor.Descriptor(
  name='FileOptions',
  full_name='google.protobuf.FileOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='java_package', full_name='google.protobuf.FileOptions.java_package', index=0,
      number=1, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='java_outer_classname', full_name='google.protobuf.FileOptions.java_outer_classname', index=1,
      number=8, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='java_multiple_files', full_name='google.protobuf.FileOptions.java_multiple_files', index=2,
      number=10, type=8, cpp_type=7, label=1,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='optimize_for', full_name='google.protobuf.FileOptions.optimize_for', index=3,
      number=9, type=14, cpp_type=8, label=1,
      default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='cc_generic_services', full_name='google.protobuf.FileOptions.cc_generic_services', index=4,
      number=16, type=8, cpp_type=7, label=1,
      default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='java_generic_services', full_name='google.protobuf.FileOptions.java_generic_services', index=5,
      number=17, type=8, cpp_type=7, label=1,
      default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='py_generic_services', full_name='google.protobuf.FileOptions.py_generic_services', index=6,
      number=18, type=8, cpp_type=7, label=1,
      default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.FileOptions.uninterpreted_option', index=7,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
    _FILEOPTIONS_OPTIMIZEMODE,
  ],
  options=None)


_MESSAGEOPTIONS = descriptor.Descriptor(
  name='MessageOptions',
  full_name='google.protobuf.MessageOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='message_set_wire_format', full_name='google.protobuf.MessageOptions.message_set_wire_format', index=0,
      number=1, type=8, cpp_type=7, label=1,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='no_standard_descriptor_accessor', full_name='google.protobuf.MessageOptions.no_standard_descriptor_accessor', index=1,
      number=2, type=8, cpp_type=7, label=1,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.MessageOptions.uninterpreted_option', index=2,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_FIELDOPTIONS = descriptor.Descriptor(
  name='FieldOptions',
  full_name='google.protobuf.FieldOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='ctype', full_name='google.protobuf.FieldOptions.ctype', index=0,
      number=1, type=14, cpp_type=8, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='packed', full_name='google.protobuf.FieldOptions.packed', index=1,
      number=2, type=8, cpp_type=7, label=1,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='deprecated', full_name='google.protobuf.FieldOptions.deprecated', index=2,
      number=3, type=8, cpp_type=7, label=1,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='experimental_map_key', full_name='google.protobuf.FieldOptions.experimental_map_key', index=3,
      number=9, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.FieldOptions.uninterpreted_option', index=4,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
    _FIELDOPTIONS_CTYPE,
  ],
  options=None)


_ENUMOPTIONS = descriptor.Descriptor(
  name='EnumOptions',
  full_name='google.protobuf.EnumOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.EnumOptions.uninterpreted_option', index=0,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_ENUMVALUEOPTIONS = descriptor.Descriptor(
  name='EnumValueOptions',
  full_name='google.protobuf.EnumValueOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.EnumValueOptions.uninterpreted_option', index=0,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_SERVICEOPTIONS = descriptor.Descriptor(
  name='ServiceOptions',
  full_name='google.protobuf.ServiceOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.ServiceOptions.uninterpreted_option', index=0,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_METHODOPTIONS = descriptor.Descriptor(
  name='MethodOptions',
  full_name='google.protobuf.MethodOptions',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='uninterpreted_option', full_name='google.protobuf.MethodOptions.uninterpreted_option', index=0,
      number=999, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_UNINTERPRETEDOPTION_NAMEPART = descriptor.Descriptor(
  name='NamePart',
  full_name='google.protobuf.UninterpretedOption.NamePart',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name_part', full_name='google.protobuf.UninterpretedOption.NamePart.name_part', index=0,
      number=1, type=9, cpp_type=9, label=2,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='is_extension', full_name='google.protobuf.UninterpretedOption.NamePart.is_extension', index=1,
      number=2, type=8, cpp_type=7, label=2,
      default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)

_UNINTERPRETEDOPTION = descriptor.Descriptor(
  name='UninterpretedOption',
  full_name='google.protobuf.UninterpretedOption',
  filename='google/protobuf/descriptor.proto',
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='name', full_name='google.protobuf.UninterpretedOption.name', index=0,
      number=2, type=11, cpp_type=10, label=3,
      default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='identifier_value', full_name='google.protobuf.UninterpretedOption.identifier_value', index=1,
      number=3, type=9, cpp_type=9, label=1,
      default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='positive_int_value', full_name='google.protobuf.UninterpretedOption.positive_int_value', index=2,
      number=4, type=4, cpp_type=4, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='negative_int_value', full_name='google.protobuf.UninterpretedOption.negative_int_value', index=3,
      number=5, type=3, cpp_type=2, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='double_value', full_name='google.protobuf.UninterpretedOption.double_value', index=4,
      number=6, type=1, cpp_type=5, label=1,
      default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='string_value', full_name='google.protobuf.UninterpretedOption.string_value', index=5,
      number=7, type=12, cpp_type=9, label=1,
      default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],  # TODO(robinson): Implement.
  enum_types=[
  ],
  options=None)


_FILEDESCRIPTORSET.fields_by_name['file'].message_type = _FILEDESCRIPTORPROTO
_FILEDESCRIPTORPROTO.fields_by_name['message_type'].message_type = _DESCRIPTORPROTO
_FILEDESCRIPTORPROTO.fields_by_name['enum_type'].message_type = _ENUMDESCRIPTORPROTO
_FILEDESCRIPTORPROTO.fields_by_name['service'].message_type = _SERVICEDESCRIPTORPROTO
_FILEDESCRIPTORPROTO.fields_by_name['extension'].message_type = _FIELDDESCRIPTORPROTO
_FILEDESCRIPTORPROTO.fields_by_name['options'].message_type = _FILEOPTIONS
_DESCRIPTORPROTO.fields_by_name['field'].message_type = _FIELDDESCRIPTORPROTO
_DESCRIPTORPROTO.fields_by_name['extension'].message_type = _FIELDDESCRIPTORPROTO
_DESCRIPTORPROTO.fields_by_name['nested_type'].message_type = _DESCRIPTORPROTO
_DESCRIPTORPROTO.fields_by_name['enum_type'].message_type = _ENUMDESCRIPTORPROTO
_DESCRIPTORPROTO.fields_by_name['extension_range'].message_type = _DESCRIPTORPROTO_EXTENSIONRANGE
_DESCRIPTORPROTO.fields_by_name['options'].message_type = _MESSAGEOPTIONS
_FIELDDESCRIPTORPROTO.fields_by_name['label'].enum_type = _FIELDDESCRIPTORPROTO_LABEL
_FIELDDESCRIPTORPROTO.fields_by_name['type'].enum_type = _FIELDDESCRIPTORPROTO_TYPE
_FIELDDESCRIPTORPROTO.fields_by_name['options'].message_type = _FIELDOPTIONS
_ENUMDESCRIPTORPROTO.fields_by_name['value'].message_type = _ENUMVALUEDESCRIPTORPROTO
_ENUMDESCRIPTORPROTO.fields_by_name['options'].message_type = _ENUMOPTIONS
_ENUMVALUEDESCRIPTORPROTO.fields_by_name['options'].message_type = _ENUMVALUEOPTIONS
_SERVICEDESCRIPTORPROTO.fields_by_name['method'].message_type = _METHODDESCRIPTORPROTO
_SERVICEDESCRIPTORPROTO.fields_by_name['options'].message_type = _SERVICEOPTIONS
_METHODDESCRIPTORPROTO.fields_by_name['options'].message_type = _METHODOPTIONS
_FILEOPTIONS.fields_by_name['optimize_for'].enum_type = _FILEOPTIONS_OPTIMIZEMODE
_FILEOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_MESSAGEOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_FIELDOPTIONS.fields_by_name['ctype'].enum_type = _FIELDOPTIONS_CTYPE
_FIELDOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_ENUMOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_ENUMVALUEOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_SERVICEOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_METHODOPTIONS.fields_by_name['uninterpreted_option'].message_type = _UNINTERPRETEDOPTION
_UNINTERPRETEDOPTION.fields_by_name['name'].message_type = _UNINTERPRETEDOPTION_NAMEPART

class FileDescriptorSet(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _FILEDESCRIPTORSET

class FileDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _FILEDESCRIPTORPROTO

class DescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class ExtensionRange(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DESCRIPTORPROTO_EXTENSIONRANGE
  DESCRIPTOR = _DESCRIPTORPROTO

class FieldDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _FIELDDESCRIPTORPROTO

class EnumDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ENUMDESCRIPTORPROTO

class EnumValueDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ENUMVALUEDESCRIPTORPROTO

class ServiceDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEDESCRIPTORPROTO

class MethodDescriptorProto(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _METHODDESCRIPTORPROTO

class FileOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _FILEOPTIONS

class MessageOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _MESSAGEOPTIONS

class FieldOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _FIELDOPTIONS

class EnumOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ENUMOPTIONS

class EnumValueOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _ENUMVALUEOPTIONS

class ServiceOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEOPTIONS

class MethodOptions(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _METHODOPTIONS

class UninterpretedOption(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class NamePart(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _UNINTERPRETEDOPTION_NAMEPART
  DESCRIPTOR = _UNINTERPRETEDOPTION




# Copyright 2010 Mike Wakerly <opensource@hoho.com>
#
# This file is part of the Pykeg package of the Kegbot project.
# For more information on Pykeg or Kegbot, see http://kegbot.org/
#
# Pykeg is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Pykeg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pykeg.  If not, see <http://www.gnu.org/licenses/>.

"""Kegweb RESTful API views."""

from functools import wraps
import hashlib
import sys

try:
  import json
except ImportError:
  try:
    import simplejson as json
  except ImportError:
    try:
      from django.utils import simplejson as json
    except ImportError:
      raise ImportError, "Unable to load a json library"

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.template import Context
from django.template.loader import get_template

from pykeg.core import backend
from pykeg.core import models
from pykeg.core import protolib
from pykeg.core import protoutil
from pykeg.web.api import common
from pykeg.web.api import forms

from google.protobuf.message import Message

INDENT = 2

### Authentication

AUTH_KEY_BASE = None
AUTH_KEY = None
if hasattr(settings, 'KEGBOT_API_AUTH_SECRET') and \
    settings.KEGBOT_API_AUTH_SECRET:
  AUTH_KEY_BASE = settings.KEGBOT_API_AUTH_SECRET
elif hasattr(settings, 'SECRET_KEY') and settings.SECRET_KEY:
  AUTH_KEY_BASE = settings.SECRET_KEY

if AUTH_KEY_BASE:
  _m = hashlib.sha256()
  _m.update(AUTH_KEY_BASE)
  AUTH_KEY = _m.hexdigest()

### Decorators

def auth_required(viewfunc):
  def _check_token(request, *args, **kwargs):
    # Check for api_auth_token; allow in either POST or GET arguments.
    tok = request.REQUEST.get('api_auth_token')
    if not tok:
      raise common.NoAuthTokenError
    if tok.lower() == AUTH_KEY.lower():
      return viewfunc(request, *args, **kwargs)
    else:
      print repr(tok.lower())
      print repr(AUTH_KEY.lower())
      raise common.BadAuthTokenError
  return wraps(viewfunc)(_check_token)

def ToJsonError(e):
  """Converts an exception to an API error response."""
  if isinstance(e, common.Error):
    code = e.__class__.__name__
    message = e.Message()
  elif isinstance(e, ValueError):
    code = 'BadRequestError'
    message = str(e)
  else:
    code = 'ServerError'
    message = 'An internal error occurred: %s' % str(e)
  result = {
    'error' : {
      'code' : code,
      'message' : message
    }
  }
  return result

def obj_to_dict(o, with_full=False):
  if hasattr(o, '__iter__'):
    return [protoutil.ProtoMessageToDict(protolib.ToProto(x, with_full)) for x in o]
  else:
    return protoutil.ProtoMessageToDict(protolib.ToProto(o, with_full))

def py_to_json(f):
  """Decorator that wraps an API method.

  The decorator transforms the method in a few ways:
    - The raw return value from the method is converted to a serialized JSON
      result.
    - The result is wrapped in an outer dict, and set as the value 'result'
    - If an exception is thrown during the method, it is converted to a protocol
      error message.
  """
  def new_function(*args, **kwargs):
    try:
      try:
        result = {'result' : f(*args, **kwargs)}
      except Http404, e:
        # We might change the HTTP status code here one day.  This also allows
        # the views to use Http404 (rather than NotFound).
        raise common.NotFoundError(e.message)
      data = json.dumps(result, indent=INDENT)
    except Exception, e:
      result = ToJsonError(e)
      data = json.dumps(result, indent=INDENT)
    return HttpResponse(data, mimetype='application/json')
  return new_function

def model_to_py(f):
  """Decorator which converts method results to native types.

  The wrapped function should return either a single Django model instance, an
  iterable of the same, or a protocol buffer Message.

  This is a bit roundabount and inefficient:  Django model instances are first
  converted to protocol buffers, then to base python types, and finally to a
  JSON-formatted string.

  TODO(mikey): revisit this layering as the API matures, or once a kegbot site
  needs to handle >0.1 qps :)
  """
  def new_function(*args, **kwargs):
    res = f(*args, **kwargs)
    request = args[0]
    with_full = False
    if request.GET.get('full') == '1':
      with_full = True
    return obj_to_dict(res, with_full)
  return new_function

def model_to_json(f):
  """Decorator that combines py_to_json and model_to_py."""
  @py_to_json
  @model_to_py
  def new_function(*args, **kwargs):
    return f(*args, **kwargs)
  return new_function

### Helpers

def _get_last_drinks(request, limit=5):
  return request.kbsite.drinks.valid()[:limit]

### Endpoints

@model_to_json
def last_drinks(request, limit=5):
  return _get_last_drinks(request, limit)

@model_to_json
def all_kegs(request):
  return request.kbsite.keg_set.all().order_by('-startdate')

@model_to_json
@auth_required   # for now, due to expense; TODO paginate me
def all_drinks(request):
  return models.Drink.objects.all().order_by('id')

@model_to_json
def get_drink(request, drink_id):
  return get_object_or_404(models.Drink, pk=drink_id, site=request.kbsite)

@model_to_json
def get_keg(request, keg_id):
  return get_object_or_404(models.Keg, pk=keg_id, site=request.kbsite)

@model_to_json
def get_keg_drinks(request, keg_id):
  keg = get_object_or_404(models.Keg, pk=keg_id, site=request.kbsite)
  return list(keg.drinks.valid())

@model_to_json
def all_taps(request):
  return request.kbsite.kegtap_set.all().order_by('name')

@model_to_json
def get_user(request, username):
  return get_object_or_404(models.User, username=username)

@model_to_json
def get_user_drinks(request, username):
  user = get_object_or_404(models.User, username=username)
  return list(user.drinks.valid())

@model_to_json
def get_auth_token(request, auth_device, token_value):
  return get_object_or_404(models.AuthenticationToken, auth_device=auth_device,
      token_value=token_value)

@model_to_json
def all_thermo_sensors(request):
  return list(request.kbsite.thermosensors.all())

def _get_sensor_or_404(request, sensor_name):
  try:
    sensor = models.ThermoSensor.objects.get(site=request.kbsite,
        raw_name=sensor_name)
  except models.ThermoSensor.DoesNotExist:
    try:
      sensor = models.ThermoSensor.objects.get(site=request.kbsite,
          nice_name=sensor_name)
    except models.ThermoSensor.DoesNotExist:
      raise Http404
  return sensor

def get_thermo_sensor(request, sensor_name):
  if request.method == 'POST':
    return thermo_sensor_post(request, sensor_name)
  else:
    return thermo_sensor_get(request, sensor_name)

@model_to_json
def thermo_sensor_get(request, sensor_name):
  sensor = _get_sensor_or_404(request, sensor_name)
  return sensor

@py_to_json
@auth_required
def thermo_sensor_post(request, sensor_name):
  sensor = _get_sensor_or_404(request, sensor_name)
  form = forms.ThermoPostForm(request.POST)
  if not form.is_valid():
    raise common.BadRequestError
  cd = form.cleaned_data
  b = backend.KegbotBackend(site=request.kbsite)
  # TODO(mikey): use form fields to compute `when`
  res = b.LogSensorReading(sensor.raw_name, cd['temp_c'])
  return protoutil.ProtoMessageToDict(res)

@model_to_json
def get_thermo_sensor_logs(request, sensor_name):
  sensor = _get_sensor_or_404(request, sensor_name)
  logs = sensor.thermolog_set.all()[:60*2]
  return logs

@py_to_json
def last_drinks_html(request, limit=5):
  last_drinks = _get_last_drinks(request, limit)

  # render each drink
  template = get_template('kegweb/drink-box.html')
  results = []
  for d in last_drinks:
    row = {}
    row['id'] = d.id
    row['box_html'] = template.render(Context({'drink': d}))
    results.append(row)
  return results

@py_to_json
def last_drink_id(request):
  last = _get_last_drinks(request, limit=1)
  if not last.count():
    return {'id': 0}
  else:
    return {'id': last[0].id}

@py_to_json
@staff_member_required
def get_access_token(request):
  return {'token': AUTH_KEY}

def tap_detail(request, tap_id):
  if request.method == 'POST':
    return tap_detail_post(request, tap_id)
  else:
    return tap_detail_get(request, tap_id)

@model_to_json
def tap_detail_get(request, tap_id):
  return get_object_or_404(models.KegTap, meter_name=tap_id, site=request.kbsite)

@py_to_json
@auth_required
def tap_detail_post(request, tap):
  form = forms.DrinkPostForm(request.POST)
  if not form.is_valid():
    raise common.BadRequestError
  cd = form.cleaned_data
  if cd.get('pour_time') and cd.get('now'):
    pour_time = datetime.datetime.fromtimestamp(cd.get('pour_time'))
    now = datetime.datetime.fromtimestamp(cd.get('now'))
    skew = datetime.datetime.now() - now
    pour_time += skew
  else:
    pour_time = None
  b = backend.KegbotBackend(site=request.kbsite)
  res = b.RecordDrink(tap_name=tap.meter_name,
    ticks=cd['ticks'],
    volume_ml=cd.get('volume_ml'),
    username=cd.get('username'),
    pour_time=pour_time,
    duration=cd.get('duration'),
    auth_token=cd.get('auth_token'))
  return protoutil.ProtoMessageToDict(res)

@py_to_json
def default_handler(request):
  raise Http404, "Not an API endpoint: %s" % request.path[:100]
